# Copyright 2018-2021 Alvaro Bartolome, alvarobartt @ GitHub
# See LICENSE for details.

import requests
from unidecode import unidecode

from .utils.constant import COUNTRY_FILTERS, FLAG_FILTERS, PAIR_FILTERS, PRODUCT_FILTERS
from .utils.extra import random_user_agent
from .utils.search_obj import SearchObj


def search_quotes(text, products=None, countries=None, n_results=None):
    """
    This function will use the Investing.com search engine so to retrieve the search results of the
    introduced text. This function will create a :obj:`list` of :obj:`investpy.utils.search_obj.SearchObj`
    class instances, unless `n_results` is set to 1, where just a single :obj:`investpy.utils.search_obj.SearchObj`
    will be returned.

    Those class instances will contain the search results so that they can be easily accessed and so
    to ease the data retrieval process since it can be done calling the methods `self.retrieve_recent_data()`
    or `self.retrieve_historical_data(from_date, to_date)` from each class instance, which will fill the historical
    data attribute, `self.data`. Also the information of the financial product can be retrieved using the
    function `self.retrieve_information()`, that will also dump the information in the attribute `self.information`;
    the technical indicators can be retrieved using `self.retrieve_technical_indicators()` dumped in the attribute
    `self.technical_indicators`; the default currency using `self.retrieve_currecy()` dumped in the attribute
    `self.default_currency`.

    Args:
        text (:obj:`str`): text to search in Investing.com among all its indexed data.
        products (:obj:`list` of :obj:`str`, optional):
            list with the product type filter/s to be applied to search result quotes so that they match
            the filters. Possible products are: `indices`, `stocks`, `etfs`, `funds`, `commodities`, `currencies`,
            `cryptos`, `bonds`, `certificates` and `fxfutures`, by default this parameter is set to `None` which
            means that no filter will be applied, and all product type quotes will be retrieved.
        countries (:obj:`list` of :obj:`str`, optional):
            list with the country name filter/s to be applied to search result quotes so that they match
            the filters. Possible countries can be found in the docs, by default this paremeter is set to
            `None` which means that no filter will be applied, and quotes from every country will be retrieved.
        n_results (:obj:`int`, optional): number of search results to retrieve and return.

    Returns:
        :obj:`list` of :obj:`investpy.utils.search_obj.SearchObj` or :obj:`investpy.utils.search_obj.SearchObj`:
            The resulting :obj:`list` of :obj:`investpy.utils.search_obj.SearchObj` will contained the retrieved
            financial products matching the introduced text if found, otherwise a RuntimeError will be raised, so as to
            let the user know that no results were found for the introduced text. But note that if the n_results value
            is equal to 1, a single value will be returned, instead of a list of values.

    Raises:
        ValueError: raised whenever any of the introduced parameter is not valid or errored.
        ConnectionError: raised whenever the connection to Investing.com failed.
        RuntimeError: raised when there was an error while executing the function.

    """

    if not text:
        raise ValueError(
            "ERR#0074: text parameter is mandatory and it should be a valid str."
        )

    if not isinstance(text, str):
        raise ValueError(
            "ERR#0074: text parameter is mandatory and it should be a valid str."
        )

    if products and not isinstance(products, list):
        raise ValueError(
            "ERR#0094: products filtering parameter is optional, but if specified, it"
            " must be a list of str."
        )

    if countries and not isinstance(countries, list):
        raise ValueError(
            "ERR#0128: countries filtering parameter is optional, but if specified, it"
            " must be a list of str."
        )

    if n_results and not isinstance(n_results, int):
        raise ValueError(
            "ERR#0088: n_results parameter is optional, but if specified, it must be an"
            " integer equal or higher than 1."
        )

    if n_results is not None:
        if n_results < 1:
            raise ValueError(
                "ERR#0088: n_results parameter is optional, but if specified, it must"
                " be an integer equal or higher than 1."
            )

    if products:
        try:
            products = list(
                map(lambda product: unidecode(product.lower().strip()), products)
            )
        except:
            raise ValueError(
                "ERR#0130: the introduced products filter must be a list of str in"
                " order to be valid."
            )

        condition = set(products).issubset(PRODUCT_FILTERS.keys())
        if condition is False:
            raise ValueError(
                'ERR#0095: products filtering parameter possible values are: "'
                + ", ".join(PRODUCT_FILTERS.keys())
                + '".'
            )

        products = [PRODUCT_FILTERS[product] for product in products]

    if countries:
        try:
            countries = list(
                map(lambda country: unidecode(country.lower().strip()), countries)
            )
        except:
            raise ValueError(
                "ERR#0131: the introduced countries filter must be a list of str in"
                " order to be valid."
            )

        condition = set(countries).issubset(COUNTRY_FILTERS.keys())
        if condition is False:
            raise ValueError(
                'ERR#0129: countries filtering parameter possible values are: "'
                + ", ".join(COUNTRY_FILTERS.keys())
                + '".'
            )

        countries = [COUNTRY_FILTERS[country] for country in countries]

    params = {
        "search_text": text,
        "tab": "quotes",
        "isFilter": True,
        "limit": 270,
        "offset": 0,
    }

    headers = {
        "User-Agent": random_user_agent(),
        "X-Requested-With": "XMLHttpRequest",
        "Accept": "text/html",
        "Accept-Encoding": "gzip, deflate",
        "Connection": "keep-alive",
    }

    url = "https://www.investing.com/search/service/SearchInnerPage"

    search_results = list()

    total_results = None

    user_limit = True if n_results is not None else False

    while True:
        req = requests.post(url, headers=headers, data=params)

        if req.status_code != 200:
            raise ConnectionError(
                f"ERR#0015: error {req.status_code}, try again later."
            )

        data = req.json()

        if data["total"]["quotes"] == 0:
            raise RuntimeError(
                "ERR#0093: no results found on Investing.com for the introduced text."
            )

        if total_results is None:
            total_results = data["total"]["quotes"]

        if n_results is None:
            n_results = data["total"]["quotes"]

        for quote in data["quotes"]:
            country, pair_type = quote["flag"], quote["pair_type"]

            if countries is not None:
                if quote["flag"] not in countries:
                    continue

            if products is not None:
                if quote["pair_type"] not in products:
                    continue

            pair_type = PAIR_FILTERS[quote["pair_type"]]

            country = None
            if pair_type not in ["cryptos", "commodities"]:
                country = (
                    FLAG_FILTERS[quote["flag"]]
                    if quote["flag"] in FLAG_FILTERS
                    else quote["flag"]
                )

            search_obj = SearchObj(
                id_=quote["pairId"],
                name=quote["name"],
                symbol=quote["symbol"],
                country=country,
                tag=quote["link"],
                pair_type=pair_type,
                exchange=quote["exchange"],
            )

            if n_results == 1 and user_limit:
                return search_obj

            if search_obj not in search_results:
                search_results.append(search_obj)

        params["offset"] += 270

        if (
            len(search_results) >= n_results
            or len(search_results) >= total_results
            or params["offset"] >= total_results
        ):
            break

    if len(search_results) < 1:
        raise RuntimeError(
            "ERR#0093: no results found on Investing.com for the introduced query."
        )

    return search_results[:n_results]


def search_events(text, importances=None, countries=None, n_results=None):
    """
    TODO
    """

    if not text:
        raise ValueError(
            "ERR#0074: text parameter is mandatory and it should be a valid str."
        )

    if not isinstance(text, str):
        raise ValueError(
            "ERR#0074: text parameter is mandatory and it should be a valid str."
        )

    if importances and not isinstance(importances, list):
        raise ValueError(
            "ERR#0138: importances filtering parameter is optional, but if specified,"
            " it must be a list of str."
        )

    if countries and not isinstance(countries, list):
        raise ValueError(
            "ERR#0128: countries filtering parameter is optional, but if specified, it"
            " must be a list of str."
        )

    if n_results and not isinstance(n_results, int):
        raise ValueError(
            "ERR#0088: n_results parameter is optional, but if specified, it must be an"
            " integer equal or higher than 1."
        )

    if n_results is not None:
        if n_results < 1:
            raise ValueError(
                "ERR#0088: n_results parameter is optional, but if specified, it must"
                " be an integer equal or higher than 1."
            )

    params = {"search_text": text, "tab": "ec_event", "limit": 270, "offset": 0}

    headers = {
        "User-Agent": random_user_agent(),
        "X-Requested-With": "XMLHttpRequest",
        "Accept": "text/html",
        "Accept-Encoding": "gzip, deflate",
        "Connection": "keep-alive",
    }

    url = "https://www.investing.com/search/service/SearchInnerPage"

    search_results = list()

    total_results = None

    while True:
        response = requests.post(url, data=params, headers=headers)

        if response.status_code != 200:
            raise ConnectionError(
                f"ERR#0015: error {response.status_code}, try again later."
            )

        events = response.json()["ec_events"]

        if len(events) == 0:
            raise RuntimeError(
                "ERR#0093: no results found on Investing.com for the introduced text."
            )

        if total_results is None:
            total_results = data["total"]["quotes"]

        if n_results is None:
            n_results = data["total"]["quotes"]

        for event in events:
            country, pair_type = quote["flag"], quote["pair_type"]

            if importances is not None:
                if quote["pair_type"] not in importances:
                    continue

            if countries is not None:
                if quote["flag"] not in countries:
                    continue

            print("TODO")
            ## pair_type = PAIR_FILTERS[quote['pair_type']]
            ## country = FLAG_FILTERS[quote['flag']]

            search_event = SearchObj(
                id_=quote["pairId"],
                name=quote["name"],
                symbol=quote["symbol"],
                country=country,
                tag=quote["link"],
                pair_type=pair_type,
                exchange=quote["exchange"],
            )

            if n_results == 1:
                return search_event

            if search_event not in search_results:
                search_results.append(search_event)

        params["offset"] += 270

        if (
            len(search_results) >= n_results
            or len(search_results) >= total_results
            or params["offset"] >= total_results
        ):
            break

    if len(search_results) < 1:
        raise RuntimeError(
            "ERR#0093: no results found on Investing.com for the introduced query."
        )

    return search_results[:n_results]



# Copyright 2018-2021 Alvaro Bartolome, alvarobartt @ GitHub
# See LICENSE for details.

import json

import pandas as pd
import pkg_resources
from unidecode import unidecode


def cryptos_as_df():
    """
    This function retrieves all the crypto data stored in `cryptos.csv` file, which previously was
    retrieved from Investing.com. Since the resulting object is a matrix of data, the crypto data is properly
    structured in rows and columns, where columns are the crypto data attribute names.

    Note that just some cryptos are available for retrieval, since Investing.com does not provide information
    from all the available ones, just the main ones.

    Returns:
        :obj:`pandas.DataFrame` - cryptos_df:
            The resulting :obj:`pandas.DataFrame` contains all the crypto data from every available crypto coin as
            indexed in Investing.com from the information previously retrieved by investpy and stored on a csv file.

            So on, the resulting :obj:`pandas.DataFrame` will look like::

                name | symbol | currency
                -----|--------|----------
                xxxx | xxxxxx | xxxxxxxx

    Raises:
        FileNotFoundError: raised if `cryptos.csv` file was not found.
        IOError: raised when `cryptos.csv` file is missing or empty.

    """

    resource_package = "investpy"
    resource_path = "/".join(("resources", "cryptos.csv"))
    if pkg_resources.resource_exists(resource_package, resource_path):
        cryptos = pd.read_csv(
            pkg_resources.resource_filename(resource_package, resource_path),
            keep_default_na=False,
        )
    else:
        raise FileNotFoundError("ERR#0081: cryptos file not found or errored.")

    if cryptos is None:
        raise IOError("ERR#0082: cryptos not found or unable to retrieve.")

    cryptos = cryptos[cryptos["status"] == "available"]

    cryptos.drop(columns=["tag", "id", "status"], inplace=True)
    cryptos = cryptos.where(pd.notnull(cryptos), None)
    cryptos.reset_index(drop=True, inplace=True)

    return cryptos


def cryptos_as_list():
    """
    This function retrieves all the crypto coin names stored in `cryptos.csv` file, which contains all the
    data from the crypto coins as previously retrieved from Investing.com. So on, this function will just return
    the crypto coin names which will be the main input parameters when it comes to crypto data retrieval functions
    from investpy.

    Note that just some cryptos are available for retrieval, since Investing.com does not provide information
    from all the available ones, just the main ones.

    Returns:
        :obj:`list` - cryptos_list:
            The resulting :obj:`list` contains the all the available crypto coin names as indexed in Investing.com
            from the information previously retrieved by investpy and stored on a csv file.

            In case the information was successfully retrieved, the :obj:`list` of crypto coin names will look like::

                cryptos_list = ['Bitcoin', 'Ethereum', 'XRP', 'Bitcoin Cash', 'Tether', 'Litecoin', ...]

    Raises:
        FileNotFoundError: raised if `cryptos.csv` file was not found.
        IOError: raised when `cryptos.csv` file is missing or empty.

    """

    resource_package = "investpy"
    resource_path = "/".join(("resources", "cryptos.csv"))
    if pkg_resources.resource_exists(resource_package, resource_path):
        cryptos = pd.read_csv(
            pkg_resources.resource_filename(resource_package, resource_path),
            keep_default_na=False,
        )
    else:
        raise FileNotFoundError("ERR#0081: cryptos file not found or errored.")

    if cryptos is None:
        raise IOError("ERR#0082: cryptos not found or unable to retrieve.")

    cryptos = cryptos[cryptos["status"] == "available"]
    cryptos.drop(columns=["tag", "id", "status"], inplace=True)
    cryptos = cryptos.where(pd.notnull(cryptos), None)

    return cryptos["name"].tolist()


def cryptos_as_dict(columns=None, as_json=False):
    """
    This function retrieves all the crypto information stored in the `cryptos.csv` file and formats it as a
    Python dictionary which contains the same information as the file, but every row is a :obj:`dict` and
    all of them are contained in a :obj:`list`. Note that the dictionary structure is the same one as the
    JSON structure. Some optional paramaters can be specified such as the columns or as_json, which are the
    column names that want to be retrieved in case of needing just some columns to avoid unnecessary information
    load, and whether the information wants to be returned as a JSON object or as a dictionary; respectively.

    Note that just some cryptos are available for retrieval, since Investing.com does not provide information
    from all the available ones, just the main ones.

    Args:
        columns (:obj:`list`, optional):column names of the crypto data to retrieve, can be: <name, currency, symbol>
        as_json (:obj:`bool`, optional): if True the returned data will be a :obj:`json` object, if False, a :obj:`list` of :obj:`dict`.

    Returns:
        :obj:`list` of :obj:`dict` OR :obj:`json` - cryptos_dict:
            The resulting :obj:`list` of :obj:`dict` contains the retrieved data from every crypto coin as indexed in Investing.com from
            the information previously retrieved by investpy and stored on a csv file.

            In case the information was successfully retrieved, the :obj:`list` of :obj:`dict` will look like::

                cryptos_dict = {
                    'name': name,
                    'currency': currency,
                    'symbol': symbol,
                }

    Raises:
        ValueError: raised whenever any of the introduced arguments is not valid.
        FileNotFoundError: raised if `cryptos.csv` file was not found.
        IOError: raised when `cryptos.csv` file is missing or empty.

    """

    if not isinstance(as_json, bool):
        raise ValueError(
            "ERR#0002: as_json argument can just be True or False, bool type."
        )

    resource_package = "investpy"
    resource_path = "/".join(("resources", "cryptos.csv"))
    if pkg_resources.resource_exists(resource_package, resource_path):
        cryptos = pd.read_csv(
            pkg_resources.resource_filename(resource_package, resource_path),
            keep_default_na=False,
        )
    else:
        raise FileNotFoundError("ERR#0081: cryptos file not found or errored.")

    if cryptos is None:
        raise IOError("ERR#0082: cryptos not found or unable to retrieve.")

    cryptos = cryptos[cryptos["status"] == "available"]
    cryptos.drop(columns=["tag", "id", "status"], inplace=True)
    cryptos = cryptos.where(pd.notnull(cryptos), None)

    if columns is None:
        columns = cryptos.columns.tolist()
    else:
        if not isinstance(columns, list):
            raise ValueError(
                "ERR#0020: specified columns argument is not a list, it can just be"
                " list type."
            )

    if not all(column in cryptos.columns.tolist() for column in columns):
        raise ValueError(
            "ERR#0021: specified columns does not exist, available columns are "
            "<name, symbol, currency>"
        )

    if as_json:
        return json.dumps(cryptos[columns].to_dict(orient="records"))
    else:
        return cryptos[columns].to_dict(orient="records")



# Copyright 2018-2021 Alvaro Bartolome, alvarobartt @ GitHub
# See LICENSE for details.

import random

import pandas as pd
import pkg_resources

from . import constant as cst


def resource_to_data(path_to_data):
    """
    This is an auxiliar function to read data from a given path, so as to wrap the load
    process of the static data files from investpy.

    Returns:
        :obj:`pandas.DataFrame` - data:
            This function returns a :obj:`pandas.DataFrame` object with all the static file's data
            retrieved from investpy.

    Raises:
        FileNotFoundError: raised if the static data file was not found.
        IOError: raised if the data file is empty or errored.

    """

    resource_package = "investpy"
    resource_path = "/".join(("resources", path_to_data))
    if pkg_resources.resource_exists(resource_package, resource_path):
        data = pd.read_csv(
            pkg_resources.resource_filename(resource_package, resource_path),
            keep_default_na=False,
        )
    else:
        raise FileNotFoundError("ERR#0115: data file not found or errored.")

    if data is None:
        raise IOError("ERR#0115: data file was empty or errored.")

    return data


def random_user_agent():
    """
    This function selects a random User-Agent from the User-Agent list, which is a constant
    variable that can be found at `investpy.utils.constant.USER_AGENTS`. User-Agents are used in
    order to avoid the limitations of the requests to Investing.com. The User-Agent is
    specified on the headers of the requests and is different for every request.

    Note that Investing.com, via changing the User-Agent on the headers of every request, allows
    a lot of requests, since it has been tested with over 10k consecutive requests without getting
    any HTTP error code from Investing.com.

    Returns:
        :obj:`str` - user_agent:
            The returned :obj:`str` is the name of a random User-Agent, which will be passed on the
            headers of a request so to avoid restrictions due to the use of multiple requests from the
            same User-Agent.

    """

    return random.choice(cst.USER_AGENTS)



# Copyright 2018-2021 Alvaro Bartolome, alvarobartt @ GitHub
# See LICENSE for details.

import json
import warnings
from datetime import date, datetime, timedelta
from random import randint

import pandas as pd
import pkg_resources
import pytz
import requests
from lxml.html import fromstring
from unidecode import unidecode

from .data.commodities_data import (
    commodities_as_df,
    commodities_as_dict,
    commodities_as_list,
    commodity_groups_list,
)
from .utils.data import Data
from .utils.extra import random_user_agent


def get_commodities(group=None):
    """
    This function retrieves all the commodities data stored in `commodities.csv` file, which previously was
    retrieved from Investing.com. Since the resulting object is a matrix of data, the commodities data is properly
    structured in rows and columns, where columns are the commodity data attribute names. Additionally, group
    filtering can be specified, so that the return commodities are from the specified group instead from every
    available group. Anyways, since it is an optional parameter it does not need to be specified, which means that
    if it is None or not specified, all the available commodities will be returned.

    Args:
        group (:obj:`str`, optional): name of the group to retrieve all the available commodities from.

    Returns:
        :obj:`pandas.DataFrame` - commodities_df:
            The resulting :obj:`pandas.DataFrame` contains all the commodities data from the introduced group if specified,
            or from all the commodity groups if None was specified, as indexed in Investing.com from the information previously
            retrieved by investpy and stored on a csv file.

            So on, the resulting :obj:`pandas.DataFrame` will look like::

                title | country | name | full_name | currency | group
                ------|---------|------|-----------|----------|-------
                xxxxx | xxxxxxx | xxxx | xxxxxxxxx | xxxxxxxx | xxxxx

    Raises:
        ValueError: raised whenever any of the introduced arguments is not valid.
        FileNotFoundError: raised when `commodities.csv` file was not found.
        IOError: raised when `commodities.csv` file is missing or empty.

    """

    return commodities_as_df(group=group)


def get_commodities_list(group=None):
    """
    This function retrieves all the commodity names as stored in `commodities.csv` file, which contains all the
    data from the commodities as previously retrieved from Investing.com. So on, this function will just return
    the commodity names from either all the available groups or from any group, which will later be used when it
    comes to both recent and historical data retrieval.

    Args:
        group (:obj:`str`, optional): name of the group to retrieve all the available commodities from.

    Returns:
        :obj:`list` - commodities_list:
            The resulting :obj:`list` contains the all the commodity names from the introduced group if specified,
            or from every group if None was specified, as indexed in Investing.com from the information previously
            retrieved by investpy and stored on a csv file.

            In case the information was successfully retrieved, the :obj:`list` of commodity names will look like::

                commodities_list = ['Gold', 'Copper', 'Silver', 'Palladium', 'Platinum', ...]

    Raises:
        ValueError: raised whenever any of the introduced arguments is not valid.
        FileNotFoundError: raised when `commodities.csv` file was not found.
        IOError: raised when `commodities.csv` file is missing or empty.

    """

    return commodities_as_list(group=group)


def get_commodities_dict(group=None, columns=None, as_json=False):
    """
    This function retrieves all the commodities information stored in the `commodities.csv` file and formats it as a
    Python dictionary which contains the same information as the file, but every row is a :obj:`dict` and
    all of them are contained in a :obj:`list`. Note that the dictionary structure is the same one as the
    JSON structure. Some optional paramaters can be specified such as the group, columns or as_json, which
    are the name of the commodity group to filter between all the available commodities so not to return all the
    commodities but just the ones from the introduced group, the column names that want to be retrieved in case
    of needing just some columns to avoid unnecessary information load, and whether the information wants to be
    returned as a JSON object or as a dictionary; respectively.

    Args:
        group (:obj:`str`, optional): name of the group to retrieve all the available commodities from.
        columns (:obj:`list`, optional):
            column names of the commodities data to retrieve, can be: <title, country, name, full_name, currency, group>
        as_json (:obj:`bool`, optional):
            if True the returned data will be a :obj:`json` object, if False, a :obj:`list` of :obj:`dict`.

    Returns:
        :obj:`list` of :obj:`dict` OR :obj:`json` - bonds_dict:
            The resulting :obj:`list` of :obj:`dict` contains the retrieved data from every bond as indexed in Investing.com from
            the information previously retrieved by investpy and stored on a csv file.

            In case the information was successfully retrieved, the :obj:`list` of :obj:`dict` will look like::

                commodities_dict = {
                    'title': title,
                    'country': country,
                    'name': name,
                    'full_name': full_name,
                    'currency': currency,
                    'group': group,
                }

    Raises:
        ValueError: raised whenever any of the introduced arguments is not valid.
        FileNotFoundError: raised when `commodities.csv` file was not found.
        IOError: raised when `commodities.csv` file is missing or empty.

    """

    return commodities_as_dict(group=group, columns=columns, as_json=as_json)


def get_commodity_groups():
    """
    This function returns a listing with all the available commodity groupsson that a filtering can be applied when
    retrieving data from commodities. The current available commodity groups are metals, agriculture and energy,
    which include all the raw materials or commodities included in them.

    Returns:
        :obj:`list` - commodity_groups:
            The resulting :obj:`list` contains all the available commodity groups as indexed in Investing.com

    Raises:
        FileNotFoundError: raised when `commodities.csv` file was not found.
        IOError: raised when `commodities.csv` file is missing or empty.

    """

    return commodity_groups_list()


def get_commodity_recent_data(
    commodity, country=None, as_json=False, order="ascending", interval="Daily"
):
    """
    This function retrieves recent historical data from the introduced commodity from Investing.com, which will be
    returned as a :obj:`pandas.DataFrame` if the parameters are valid and the request to Investing.com succeeds.
    Note that additionally some optional parameters can be specified: as_json and order, which let the user decide
    if the data is going to be returned as a :obj:`json` or not, and if the historical data is going to be ordered
    ascending or descending (where the index is the date), respectively.

    Args:
        commodity (:obj:`str`): name of the commodity to retrieve recent data from.
        country (:obj:`str`, optional):
            name of the country to retrieve the commodity data from (if there is more than one country that
            provides data from the same commodity).
        as_json (:obj:`bool`, optional):
            to determine the format of the output data, either a :obj:`pandas.DataFrame` if False and a :obj:`json` if True.
        order (:obj:`str`, optional): to define the order of the retrieved data which can either be ascending or descending.
        interval (:obj:`str`, optional):
            value to define the historical data interval to retrieve, by default `Daily`, but it can also be `Weekly` or `Monthly`.

    Returns:
        :obj:`pandas.DataFrame` or :obj:`json`:
            The function can return either a :obj:`pandas.DataFrame` or a :obj:`json` object, containing the retrieved
            recent data of the specified commodity. So on, the resulting dataframe contains the open, high, low and close
            values for the selected commodity on market days and the currency in which those values are presented.

            The returned data is case we use default arguments will look like::

                Date || Open | High | Low | Close | Volume | Currency
                -----||------|------|-----|-------|--------|----------
                xxxx || xxxx | xxxx | xxx | xxxxx | xxxxxx | xxxxxxxx

            but in case that as_json parameter was defined as True, then the output will be::

                {
                    name: name,
                    recent: [
                        {
                            date: 'dd/mm/yyyy',
                            open: x,
                            high: x,
                            low: x,
                            close: x,
                            volume: x,
                            currency: x
                        },
                        ...
                    ]
                }

    Raises:
        ValueError: raised whenever any of the introduced arguments is not valid or errored.
        IOError: raised if commodities object/file was not found or unable to retrieve.
        RuntimeError: raised if the introduced commodity was not found or did not match any of the existing ones.
        ConnectionError: raised if connection to Investing.com could not be established.
        IndexError: raised if commodity recent data was unavailable or not found in Investing.com.

    Examples:
        >>> data = investpy.get_commodity_recent_data(commodity='gold')
        >>> data.head()
                      Open    High     Low   Close  Volume Currency
        Date
        2019-10-25  1506.4  1520.9  1503.1  1505.3  368743      USD
        2019-10-28  1507.4  1510.8  1492.3  1495.8  318126      USD
        2019-10-29  1494.3  1497.1  1485.6  1490.7  291980      USD
        2019-10-30  1490.5  1499.3  1483.1  1496.7  353638      USD
        2019-10-31  1498.8  1516.7  1496.0  1514.8  390013      USD

    """

    if not commodity:
        raise ValueError(
            "ERR#0078: commodity parameter is mandatory and must be a valid commodity"
            " name."
        )

    if not isinstance(commodity, str):
        raise ValueError(
            "ERR#0078: commodity parameter is mandatory and must be a valid commodity"
            " name."
        )

    if country is not None and not isinstance(country, str):
        raise ValueError("ERR#0025: specified country value not valid.")

    if not isinstance(as_json, bool):
        raise ValueError(
            "ERR#0002: as_json argument can just be True or False, bool type."
        )

    if order not in ["ascending", "asc", "descending", "desc"]:
        raise ValueError(
            "ERR#0003: order argument can just be ascending (asc) or descending (desc),"
            " str type."
        )

    if not interval:
        raise ValueError(
            "ERR#0073: interval value should be a str type and it can just be either"
            " 'Daily', 'Weekly' or 'Monthly'."
        )

    if not isinstance(interval, str):
        raise ValueError(
            "ERR#0073: interval value should be a str type and it can just be either"
            " 'Daily', 'Weekly' or 'Monthly'."
        )

    interval = interval.lower()

    if interval not in ["daily", "weekly", "monthly"]:
        raise ValueError(
            "ERR#0073: interval value should be a str type and it can just be either"
            " 'Daily', 'Weekly' or 'Monthly'."
        )

    resource_package = "investpy"
    resource_path = "/".join(("resources", "commodities.csv"))
    if pkg_resources.resource_exists(resource_package, resource_path):
        commodities = pd.read_csv(
            pkg_resources.resource_filename(resource_package, resource_path),
            keep_default_na=False,
        )
    else:
        raise FileNotFoundError("ERR#0075: commodities file not found or errored.")

    if commodities is None:
        raise IOError("ERR#0076: commodities not found or unable to retrieve.")

    commodity = unidecode(commodity.strip().lower())

    if commodity not in list(commodities["name"].apply(unidecode).str.lower()):
        raise RuntimeError(
            "ERR#0079: commodity " + commodity + " not found, check if it is correct."
        )

    if country is None:
        found_commodities = commodities[
            commodities["name"].apply(unidecode).str.lower() == commodity
        ]

        if len(found_commodities) > 1:
            msg = (
                "Note that the displayed commodity data can differ depending on the"
                " country. If you want to retrieve "
                + commodity
                + " data from either "
                + " or ".join(found_commodities["country"].tolist())
                + ", specify the country parameter."
            )
            warnings.warn(msg, Warning)

        del found_commodities
    else:
        country = unidecode(country.strip().lower())

        if country not in list(set(commodities["country"].str.lower())):
            raise RuntimeError(
                "ERR#0034: country " + country + " not found, check if it is correct."
            )

        commodities = commodities[commodities["country"] == country]

    full_name = commodities.loc[
        (commodities["name"].apply(unidecode).str.lower() == commodity).idxmax(),
        "full_name",
    ]
    id_ = commodities.loc[
        (commodities["name"].apply(unidecode).str.lower() == commodity).idxmax(), "id"
    ]
    name = commodities.loc[
        (commodities["name"].apply(unidecode).str.lower() == commodity).idxmax(), "name"
    ]

    currency = commodities.loc[
        (commodities["name"].apply(unidecode).str.lower() == commodity).idxmax(),
        "currency",
    ]

    header = full_name + " Historical Data"

    params = {
        "curr_id": id_,
        "smlID": str(randint(1000000, 99999999)),
        "header": header,
        "interval_sec": interval.capitalize(),
        "sort_col": "date",
        "sort_ord": "DESC",
        "action": "historical_data",
    }

    head = {
        "User-Agent": random_user_agent(),
        "X-Requested-With": "XMLHttpRequest",
        "Accept": "text/html",
        "Accept-Encoding": "gzip, deflate",
        "Connection": "keep-alive",
    }

    url = "https://www.investing.com/instruments/HistoricalDataAjax"

    req = requests.post(url, headers=head, data=params)

    if req.status_code != 200:
        raise ConnectionError(
            "ERR#0015: error " + str(req.status_code) + ", try again later."
        )

    root_ = fromstring(req.text)
    path_ = root_.xpath(".//table[@id='curr_table']/tbody/tr")
    result = list()

    if path_:
        for elements_ in path_:
            if elements_.xpath(".//td")[0].text_content() == "No results found":
                raise IndexError(
                    "ERR#0080: commodity information unavailable or not found."
                )

            info = []

            for nested_ in elements_.xpath(".//td"):
                info.append(nested_.get("data-real-value"))

            commodity_date = datetime.strptime(
                str(
                    datetime.fromtimestamp(int(info[0]), tz=pytz.timezone("GMT")).date()
                ),
                "%Y-%m-%d",
            )

            commodity_close = float(info[1].replace(",", ""))
            commodity_open = float(info[2].replace(",", ""))
            commodity_high = float(info[3].replace(",", ""))
            commodity_low = float(info[4].replace(",", ""))

            commodity_volume = int(info[5])

            result.insert(
                len(result),
                Data(
                    commodity_date,
                    commodity_open,
                    commodity_high,
                    commodity_low,
                    commodity_close,
                    commodity_volume,
                    currency,
                    None,
                ),
            )

        if order in ["ascending", "asc"]:
            result = result[::-1]
        elif order in ["descending", "desc"]:
            result = result

        if as_json is True:
            json_ = {
                "name": name,
                "recent": [value.commodity_as_json() for value in result],
            }

            return json.dumps(json_, sort_keys=False)
        elif as_json is False:
            df = pd.DataFrame.from_records(
                [value.commodity_to_dict() for value in result]
            )
            df.set_index("Date", inplace=True)

            return df
    else:
        raise RuntimeError("ERR#0004: data retrieval error while scraping.")


def get_commodity_historical_data(
    commodity,
    from_date,
    to_date,
    country=None,
    as_json=False,
    order="ascending",
    interval="Daily",
):
    """
    This function retrieves historical data from the introduced commodity from Investing.com. So on, the historical data
    of the introduced commodity in the specified date range will be retrieved and returned as a :obj:`pandas.DataFrame`
    if the parameters are valid and the request to Investing.com succeeds. Note that additionally some optional parameters
    can be specified: as_json and order, which let the user decide if the data is going to be returned as a :obj:`json` or not,
    and if the historical data is going to be ordered ascending or descending (where the index is the date), respectively.

    Args:
        commodity (:obj:`str`): name of the commodity to retrieve recent data from.
        from_date (:obj:`str`): date formatted as `dd/mm/yyyy`, since when data is going to be retrieved.
        to_date (:obj:`str`): date formatted as `dd/mm/yyyy`, until when data is going to be retrieved.
        country (:obj:`str`, optional):
            name of the country to retrieve the commodity data from (if there is more than one country that
            provides data from the same commodity).
        as_json (:obj:`bool`, optional):
            to determine the format of the output data, either a :obj:`pandas.DataFrame` if False and a :obj:`json` if True.
        order (:obj:`str`, optional): to define the order of the retrieved data which can either be ascending or descending.
        interval (:obj:`str`, optional):
            value to define the historical data interval to retrieve, by default `Daily`, but it can also be `Weekly` or `Monthly`.

    Returns:
        :obj:`pandas.DataFrame` or :obj:`json`:
            The function returns a either a :obj:`pandas.DataFrame` or a :obj:`json` file containing the retrieved
            historical data of the specified commodity. So on, the resulting dataframe contains the open, high, low and close
            values for the selected commodity on market days and the currency in which those values are presented.

            The returned data is case we use default arguments will look like::

                Date || Open | High | Low | Close | Volume | Currency
                -----||------|------|-----|-------|--------|----------
                xxxx || xxxx | xxxx | xxx | xxxxx | xxxxxx | xxxxxxxx

            but in case that as_json parameter was defined as True, then the output will be::

                {
                    name: name,
                    historical: [
                        {
                            date: 'dd/mm/yyyy',
                            open: x,
                            high: x,
                            low: x,
                            close: x,
                            volume: x,
                            currency: x
                        },
                        ...
                    ]
                }

    Raises:
        ValueError: raised whenever any of the introduced arguments is not valid or errored.
        IOError: raised if commodities object/file was not found or unable to retrieve.
        RuntimeError: raised if the introduced commodity was not found or did not match any of the existing ones.
        ConnectionError: raised if connection to Investing.com could not be established.
        IndexError: raised if commodity historical data was unavailable or not found in Investing.com.

    Examples:
        >>> data = investpy.get_commodity_historical_data(commodity='gold', from_date='01/01/2018', to_date='01/01/2019')
        >>> data.head()
                      Open    High     Low   Close  Volume Currency
        Date
        2018-01-01  1305.8  1309.7  1304.6  1308.7       0      USD
        2018-01-02  1370.5  1370.5  1370.5  1370.5      97      USD
        2018-01-03  1372.0  1372.0  1369.0  1374.2      22      USD
        2018-01-04  1363.4  1375.6  1362.7  1377.4      13      USD
        2018-01-05  1377.8  1377.8  1377.8  1378.4      10      USD

    """

    if not commodity:
        raise ValueError(
            "ERR#0078: commodity parameter is mandatory and must be a valid commodity"
            " name."
        )

    if not isinstance(commodity, str):
        raise ValueError(
            "ERR#0078: commodity parameter is mandatory and must be a valid commodity"
            " name."
        )

    if country is not None and not isinstance(country, str):
        raise ValueError("ERR#0025: specified country value not valid.")

    if not isinstance(as_json, bool):
        raise ValueError(
            "ERR#0002: as_json argument can just be True or False, bool type."
        )

    if order not in ["ascending", "asc", "descending", "desc"]:
        raise ValueError(
            "ERR#0003: order argument can just be ascending (asc) or descending (desc),"
            " str type."
        )

    if not interval:
        raise ValueError(
            "ERR#0073: interval value should be a str type and it can just be either"
            " 'Daily', 'Weekly' or 'Monthly'."
        )

    if not isinstance(interval, str):
        raise ValueError(
            "ERR#0073: interval value should be a str type and it can just be either"
            " 'Daily', 'Weekly' or 'Monthly'."
        )

    interval = interval.lower()

    if interval not in ["daily", "weekly", "monthly"]:
        raise ValueError(
            "ERR#0073: interval value should be a str type and it can just be either"
            " 'Daily', 'Weekly' or 'Monthly'."
        )

    try:
        datetime.strptime(from_date, "%d/%m/%Y")
    except ValueError:
        raise ValueError(
            "ERR#0011: incorrect from_date date format, it should be 'dd/mm/yyyy'."
        )

    try:
        datetime.strptime(to_date, "%d/%m/%Y")
    except ValueError:
        raise ValueError(
            "ERR#0012: incorrect to_date format, it should be 'dd/mm/yyyy'."
        )

    start_date = datetime.strptime(from_date, "%d/%m/%Y")
    end_date = datetime.strptime(to_date, "%d/%m/%Y")

    if start_date >= end_date:
        raise ValueError(
            "ERR#0032: to_date should be greater than from_date, both formatted as"
            " 'dd/mm/yyyy'."
        )

    date_interval = {
        "intervals": [],
    }

    flag = True

    while flag is True:
        diff = end_date.year - start_date.year

        if diff > 19:
            obj = {
                "start": start_date.strftime("%m/%d/%Y"),
                "end": start_date.replace(year=start_date.year + 19).strftime(
                    "%m/%d/%Y"
                ),
            }

            date_interval["intervals"].append(obj)

            start_date = start_date.replace(year=start_date.year + 19) + timedelta(
                days=1
            )
        else:
            obj = {
                "start": start_date.strftime("%m/%d/%Y"),
                "end": end_date.strftime("%m/%d/%Y"),
            }

            date_interval["intervals"].append(obj)

            flag = False

    interval_limit = len(date_interval["intervals"])
    interval_counter = 0

    data_flag = False

    resource_package = "investpy"
    resource_path = "/".join(("resources", "commodities.csv"))
    if pkg_resources.resource_exists(resource_package, resource_path):
        commodities = pd.read_csv(
            pkg_resources.resource_filename(resource_package, resource_path),
            keep_default_na=False,
        )
    else:
        raise FileNotFoundError("ERR#0075: commodities file not found or errored.")

    if commodities is None:
        raise IOError("ERR#0076: commodities not found or unable to retrieve.")

    commodity = unidecode(commodity.strip().lower())

    if commodity not in list(commodities["name"].apply(unidecode).str.lower()):
        raise RuntimeError(
            "ERR#0079: commodity " + commodity + " not found, check if it is correct."
        )

    if country is None:
        found_commodities = commodities[
            commodities["name"].apply(unidecode).str.lower() == commodity
        ]

        if len(found_commodities) > 1:
            msg = (
                "Note that the displayed commodity data can differ depending on the"
                " country. If you want to retrieve "
                + commodity
                + " data from either "
                + " or ".join(found_commodities["country"].tolist())
                + ", specify the country parameter."
            )
            warnings.warn(msg, Warning)

        del found_commodities
    else:
        country = unidecode(country.strip().lower())

        if country not in list(set(commodities["country"].str.lower())):
            raise RuntimeError(
                "ERR#0034: country " + country + " not found, check if it is correct."
            )

        commodities = commodities[commodities["country"] == country]

    full_name = commodities.loc[
        (commodities["name"].apply(unidecode).str.lower() == commodity).idxmax(),
        "full_name",
    ]
    id_ = commodities.loc[
        (commodities["name"].apply(unidecode).str.lower() == commodity).idxmax(), "id"
    ]
    name = commodities.loc[
        (commodities["name"].apply(unidecode).str.lower() == commodity).idxmax(), "name"
    ]

    currency = commodities.loc[
        (commodities["name"].apply(unidecode).str.lower() == commodity).idxmax(),
        "currency",
    ]

    header = full_name + " Historical Data"

    final = list()

    for index in range(len(date_interval["intervals"])):
        interval_counter += 1

        params = {
            "curr_id": id_,
            "smlID": str(randint(1000000, 99999999)),
            "header": header,
            "st_date": date_interval["intervals"][index]["start"],
            "end_date": date_interval["intervals"][index]["end"],
            "interval_sec": interval.capitalize(),
            "sort_col": "date",
            "sort_ord": "DESC",
            "action": "historical_data",
        }

        head = {
            "User-Agent": random_user_agent(),
            "X-Requested-With": "XMLHttpRequest",
            "Accept": "text/html",
            "Accept-Encoding": "gzip, deflate",
            "Connection": "keep-alive",
        }

        url = "https://www.investing.com/instruments/HistoricalDataAjax"

        req = requests.post(url, headers=head, data=params)

        if req.status_code != 200:
            raise ConnectionError(
                "ERR#0015: error " + str(req.status_code) + ", try again later."
            )

        if not req.text:
            continue

        root_ = fromstring(req.text)
        path_ = root_.xpath(".//table[@id='curr_table']/tbody/tr")

        result = list()

        if path_:
            for elements_ in path_:
                if elements_.xpath(".//td")[0].text_content() == "No results found":
                    if interval_counter < interval_limit:
                        data_flag = False
                    else:
                        raise IndexError(
                            "ERR#0080: commodity information unavailable or not found."
                        )
                else:
                    data_flag = True

                info = []

                for nested_ in elements_.xpath(".//td"):
                    info.append(nested_.get("data-real-value"))

                if data_flag is True:
                    commodity_date = datetime.strptime(
                        str(
                            datetime.fromtimestamp(
                                int(info[0]), tz=pytz.timezone("GMT")
                            ).date()
                        ),
                        "%Y-%m-%d",
                    )

                    commodity_close = float(info[1].replace(",", ""))
                    commodity_open = float(info[2].replace(",", ""))
                    commodity_high = float(info[3].replace(",", ""))
                    commodity_low = float(info[4].replace(",", ""))

                    commodity_volume = int(info[5])

                    result.insert(
                        len(result),
                        Data(
                            commodity_date,
                            commodity_open,
                            commodity_high,
                            commodity_low,
                            commodity_close,
                            commodity_volume,
                            currency,
                            None,
                        ),
                    )

            if data_flag is True:
                if order in ["ascending", "asc"]:
                    result = result[::-1]
                elif order in ["descending", "desc"]:
                    result = result

                if as_json is True:
                    json_list = [value.commodity_as_json() for value in result]

                    final.append(json_list)
                elif as_json is False:
                    df = pd.DataFrame.from_records(
                        [value.commodity_to_dict() for value in result]
                    )
                    df.set_index("Date", inplace=True)

                    final.append(df)

        else:
            raise RuntimeError("ERR#0004: data retrieval error while scraping.")

    if order in ["descending", "desc"]:
        final.reverse()

    if as_json is True:
        json_ = {
            "name": name,
            "historical": [value for json_list in final for value in json_list],
        }
        return json.dumps(json_, sort_keys=False)
    elif as_json is False:
        return pd.concat(final)


def get_commodity_information(commodity, country=None, as_json=False):
    """
    This function retrieves fundamental financial information from the specified commodity. The retrieved
    information from the commodity can be valuable as it is additional information that can be used combined
    with OHLC values, so to determine financial insights from the company which holds the specified commodity.

    Args:
        commodity (:obj:`str`): name of the commodity to retrieve information from.
        country (:obj:`str`, optional):
            name of the country to retrieve the commodity information from (if there is more than one country
            that provides data from the same commodity).
        as_json (:obj:`bool`, optional):
            optional argument to determine the format of the output data (:obj:`dict` or :obj:`json`).

    Returns:
        :obj:`pandas.DataFrame` or :obj:`dict`- commodity_information:
            The resulting :obj:`pandas.DataFrame` contains the information fields retrieved from Investing.com
            from the specified commodity; it can also be returned as a :obj:`dict`, if argument `as_json=True`.

            If any of the information fields could not be retrieved, that field/s will be filled with
            None values. If the retrieval process succeeded, the resulting :obj:`dict` will look like::

                commodity_information = {
                    "1-Year Change": "16.15%",
                    "52 wk Range": "1,270.2 - 1,566.2",
                    "Base Symbol": "GC",
                    "Commodity Name": "Gold",
                    "Contract Size": "100 Troy Ounces",
                    "Last Rollover Day": "24/11/2019",
                    "Month": "Feb 20",
                    "Months": "GJMQVZ",
                    "Open": 1479.8,
                    "Point Value": "$100",
                    "Prev. Close": 1481.2,
                    "Settlement Day": "25/01/2020",
                    "Settlement Type": "Physical",
                    "Tick Size": 0.1,
                    "Tick Value": 10.0,
                    "Day's Range": "1,477.55 - 1,484.25"
                }

    Raises:
        ValueError: raised if any of the introduced arguments is not valid or errored.
        FileNotFoundError: raised if `commodities.csv` file was not found or errored.
        IOError: raised if commodities.csv file is empty or errored.
        RuntimeError: raised if scraping process failed while running.
        ConnectionError: raised if the connection to Investing.com errored (did not return HTTP 200)

    """

    if not commodity:
        raise ValueError(
            "ERR#0078: commodity parameter is mandatory and must be a valid commodity"
            " name."
        )

    if not isinstance(commodity, str):
        raise ValueError(
            "ERR#0078: commodity parameter is mandatory and must be a valid commodity"
            " name."
        )

    if country is not None and not isinstance(country, str):
        raise ValueError("ERR#0025: specified country value not valid.")

    if not isinstance(as_json, bool):
        raise ValueError(
            "ERR#0002: as_json argument can just be True or False, bool type."
        )

    resource_package = "investpy"
    resource_path = "/".join(("resources", "commodities.csv"))
    if pkg_resources.resource_exists(resource_package, resource_path):
        commodities = pd.read_csv(
            pkg_resources.resource_filename(resource_package, resource_path),
            keep_default_na=False,
        )
    else:
        raise FileNotFoundError("ERR#0075: commodities file not found or errored.")

    if commodities is None:
        raise IOError("ERR#0076: commodities not found or unable to retrieve.")

    commodity = unidecode(commodity.strip().lower())

    if commodity not in list(commodities["name"].apply(unidecode).str.lower()):
        raise RuntimeError(
            "ERR#0079: commodity " + commodity + " not found, check if it is correct."
        )

    if country is None:
        found_commodities = commodities[
            commodities["name"].apply(unidecode).str.lower() == commodity
        ]

        if len(found_commodities) > 1:
            msg = (
                "Note that the displayed commodity information can differ depending on"
                " the country. If you want to retrieve "
                + commodity
                + " data from either "
                + " or ".join(found_commodities["country"].tolist())
                + ", specify the country parameter."
            )
            warnings.warn(msg, Warning)

        del found_commodities
    else:
        country = unidecode(country.strip().lower())

        if country not in list(set(commodities["country"].str.lower())):
            raise RuntimeError(
                "ERR#0034: country " + country + " not found, check if it is correct."
            )

        commodities = commodities[commodities["country"] == country]

    name = commodities.loc[
        (commodities["name"].apply(unidecode).str.lower() == commodity).idxmax(), "name"
    ]
    tag = commodities.loc[
        (commodities["name"].apply(unidecode).str.lower() == commodity).idxmax(), "tag"
    ]

    url = "https://www.investing.com/commodities/" + tag

    head = {
        "User-Agent": random_user_agent(),
        "X-Requested-With": "XMLHttpRequest",
        "Accept": "text/html",
        "Accept-Encoding": "gzip, deflate",
        "Connection": "keep-alive",
    }

    req = requests.get(url, headers=head)

    if req.status_code != 200:
        raise ConnectionError(
            "ERR#0015: error " + str(req.status_code) + ", try again later."
        )

    root_ = fromstring(req.text)
    path_ = root_.xpath("//dl[@data-test='key-info']/div")

    result = pd.DataFrame(
        columns=[
            "Commodity Name",
            "Prev. Close",
            "Month",
            "Tick Size",
            "Open",
            "Contract Size",
            "Tick Value",
            "Day's Range",
            "Settlement Type",
            "Base Symbol",
            "52 wk Range",
            "Settlement Day",
            "Point Value",
            "1-Year Change",
            "Last Rollover Day",
            "Months",
        ]
    )
    result.at[0, "Commodity Name"] = name

    if not path_:
        raise RuntimeError("ERR#0004: data retrieval error while scraping.")

    for elements_ in path_:
        title_ = elements_.xpath(".//dt")[0].text_content()
        element = elements_.xpath(".//dd")[0]
        if title_ in result.columns.tolist():
            try:
                result.at[0, title_] = float(element.text_content().replace(",", ""))
                continue
            except:
                pass
            try:
                text = element.text_content().strip()
                result.at[0, title_] = datetime.strptime(text, "%m/%d/%Y").strftime(
                    "%d/%m/%Y"
                )
                continue
            except:
                pass
            try:
                text = element.text_content().strip()
                if text.__contains__("1 = "):
                    result.at[0, title_] = text.replace("1 = ", "")
                    continue
            except:
                pass
            try:
                value = element.text_content().strip()
                if value.__contains__("K"):
                    value = float(value.replace("K", "").replace(",", "")) * 1e3
                elif value.__contains__("M"):
                    value = float(value.replace("M", "").replace(",", "")) * 1e6
                elif value.__contains__("B"):
                    value = float(value.replace("B", "").replace(",", "")) * 1e9
                elif value.__contains__("T"):
                    value = float(value.replace("T", "").replace(",", "")) * 1e12
                result.at[0, title_] = value
                continue
            except:
                pass

    result.replace({"N/A": None}, inplace=True)

    if as_json is True:
        json_ = result.iloc[0].to_dict()
        return json_
    elif as_json is False:
        return result


def get_commodities_overview(group, as_json=False, n_results=100):
    """
    This function retrieves an overview containing all the real time data available for the main commodities from
    every commodity group (metals, softs, meats, energy and grains), such as the names, symbols, current value, etc.
    as indexed in Investing.com. So on, the main usage of this function is to get an overview on the main commodities
    from a group, so to get a general view. Note that since this function is retrieving a lot of information at once,
    by default just the overview of the Top 100 commodities is being retrieved, but an additional parameter called n_results
    can be specified so to retrieve N results. Anyways, note that in commodities case, there are just a few ones available.

    Args:
        group (:obj:`str`): name of the commodity group to retrieve an overview from.
        as_json (:obj:`bool`, optional):
            optional argument to determine the format of the output data (:obj:`pandas.DataFrame` or :obj:`json`).
        n_results (:obj:`int`, optional): number of results to be displayed on the overview table (0-1000).

    Returns:
        :obj:`pandas.DataFrame` - commodities_overview:
            The resulting :obj:`pandas.DataFrame` contains all the data available in Investing.com of the main commodities
            from a commodity group in order to get an overview of it.

            If the retrieval process succeeded, the resulting :obj:`pandas.DataFrame` should look like::

                country | name | last | last_close | high | low | change | change_percentage | currency
                --------|------|------|------------|------|-----|--------|-------------------|----------
                xxxxxxx | xxxx | xxxx | xxxxxxxxxx | xxxx | xxx | xxxxxx | xxxxxxxxxxxxxxxxx | xxxxxxxx

    Raises:
        ValueError: raised if any of the introduced arguments errored.
        FileNotFoundError: raised if `commodities.csv` file is missing.
        IOError: raised if data could not be retrieved due to file error.
        RuntimeError:
            raised either if the introduced group does not match any of the listed ones or if no overview results could be
            retrieved from Investing.com.
        ConnectionError: raised if GET requests does not return 200 status code.

    """

    if group is None:
        raise ValueError("ERR#0090: group can not be None, it should be a str.")

    if group is not None and not isinstance(group, str):
        raise ValueError("ERR#0090: group can not be None, it should be a str.")

    if not isinstance(as_json, bool):
        raise ValueError(
            "ERR#0002: as_json argument can just be True or False, bool type."
        )

    if not isinstance(n_results, int):
        raise ValueError(
            "ERR#0089: n_results argument should be an integer between 1 and 1000."
        )

    if 1 > n_results or n_results > 1000:
        raise ValueError(
            "ERR#0089: n_results argument should be an integer between 1 and 1000."
        )

    resource_package = "investpy"
    resource_path = "/".join(("resources", "commodities.csv"))
    if pkg_resources.resource_exists(resource_package, resource_path):
        commodities = pd.read_csv(
            pkg_resources.resource_filename(resource_package, resource_path),
            keep_default_na=False,
        )
    else:
        raise FileNotFoundError("ERR#0075: commodities file not found or errored.")

    if commodities is None:
        raise IOError("ERR#0076: commodities not found or unable to retrieve.")

    group = unidecode(group.strip().lower())

    if group not in get_commodity_groups():
        raise RuntimeError("ERR#0091: specified commodity group value is not valid.")

    commodities = commodities[commodities["group"] == group]

    head = {
        "User-Agent": random_user_agent(),
        "X-Requested-With": "XMLHttpRequest",
        "Accept": "text/html",
        "Accept-Encoding": "gzip, deflate",
        "Connection": "keep-alive",
    }

    url = "https://www.investing.com/commodities/" + group

    req = requests.get(url, headers=head)

    if req.status_code != 200:
        raise ConnectionError(
            "ERR#0015: error " + str(req.status_code) + ", try again later."
        )

    root_ = fromstring(req.text)
    table = root_.xpath(".//table[@id='cross_rate_1']/tbody/tr")

    results = list()

    if len(table) > 0:
        for row in table[:n_results]:
            id_ = row.get("id").replace("pair_", "")
            country_check = (
                row.xpath(".//td[@class='flag']/span")[0].get("title").lower()
            )

            name = (
                row.xpath(".//td[contains(@class, 'elp')]/a")[0].text_content().strip()
            )

            pid = "pid-" + id_

            last = row.xpath(".//td[@class='" + pid + "-last']")[0].text_content()
            last_close = row.xpath(".//td[@class='" + pid + "-last_close']")[
                0
            ].text_content()
            high = row.xpath(".//td[@class='" + pid + "-high']")[0].text_content()
            low = row.xpath(".//td[@class='" + pid + "-low']")[0].text_content()

            pc = row.xpath(".//td[contains(@class, '" + pid + "-pc')]")[
                0
            ].text_content()
            pcp = row.xpath(".//td[contains(@class, '" + pid + "-pcp')]")[
                0
            ].text_content()

            data = {
                "country": country_check if country_check != "" else None,
                "name": name,
                "last": float(last.replace(",", "")),
                "last_close": float(last_close.replace(",", "")),
                "high": float(high.replace(",", "")),
                "low": float(low.replace(",", "")),
                "change": pc,
                "change_percentage": pcp,
                "currency": commodities.loc[
                    (
                        (commodities["name"] == name)
                        & (commodities["country"] == country_check)
                    ).idxmax(),
                    "currency",
                ]
                if country_check != ""
                else commodities.loc[
                    (commodities["name"] == name).idxmax(), "currency"
                ],
            }

            results.append(data)
    else:
        raise RuntimeError(
            "ERR#0092: no data found while retrieving the overview from Investing.com"
        )

    df = pd.DataFrame(results)

    if as_json:
        return json.loads(df.to_json(orient="records"))
    else:
        return df


def search_commodities(by, value):
    """
    This function searches commodities by the introduced value for the specified field. This means that this function
    is going to search if there is a value that matches the introduced one for the specified field which is the
    `commodities.csv` column name to search in. Available fields to search commodities are 'name', 'full_name' and 'title'.

    Args:
        by (:obj:`str`): name of the field to search for, which is the column name which can be: ''name', 'full_name' or 'title'.
        value (:obj:`str`): value of the field to search for, which is the value that is going to be searched.

    Returns:
        :obj:`pandas.DataFrame` - search_result:
            The resulting :obj:`pandas.DataFrame` contains the search results from the given query, which is
            any match of the specified value in the specified field. If there are no results for the given query,
            an error will be raised, but otherwise the resulting :obj:`pandas.DataFrame` will contain all the
            available commodities that match the introduced query.

    Raises:
        ValueError: raised if any of the introduced parameters is not valid or errored.
        FileNotFoundError: raised if `commodities.csv` file is missing.
        IOError: raised if data could not be retrieved due to file error.
        RuntimeError: raised if no results were found for the introduced value in the introduced field.

    """

    if not by:
        raise ValueError(
            "ERR#0006: the introduced field to search is mandatory and should be a str."
        )

    if not isinstance(by, str):
        raise ValueError(
            "ERR#0006: the introduced field to search is mandatory and should be a str."
        )

    if not value:
        raise ValueError(
            "ERR#0017: the introduced value to search is mandatory and should be a str."
        )

    if not isinstance(value, str):
        raise ValueError(
            "ERR#0017: the introduced value to search is mandatory and should be a str."
        )

    resource_package = "investpy"
    resource_path = "/".join(("resources", "commodities.csv"))
    if pkg_resources.resource_exists(resource_package, resource_path):
        commodities = pd.read_csv(
            pkg_resources.resource_filename(resource_package, resource_path),
            keep_default_na=False,
        )
    else:
        raise FileNotFoundError("ERR#0075: commodities file not found or errored.")

    if commodities is None:
        raise IOError("ERR#0076: commodities not found or unable to retrieve.")

    commodities.drop(columns=["tag", "id"], inplace=True)

    available_search_fields = commodities.columns.tolist()

    if isinstance(by, str) and by not in available_search_fields:
        raise ValueError(
            "ERR#0026: the introduced field to search can either just be "
            + " or ".join(available_search_fields)
        )

    commodities["matches"] = commodities[by].str.contains(value, case=False)

    search_result = commodities.loc[commodities["matches"] == True].copy()

    if len(search_result) == 0:
        raise RuntimeError(
            "ERR#0043: no results were found for the introduced " + str(by) + "."
        )

    search_result.drop(columns=["matches"], inplace=True)
    search_result.reset_index(drop=True, inplace=True)

    return search_result



# Copyright 2018-2021 Alvaro Bartolome, alvarobartt @ GitHub
# See LICENSE for details.

import json
import re
from datetime import date, datetime, timedelta
from random import randint

import pandas as pd
import pkg_resources
import pytz
import requests
from lxml.html import fromstring
from unidecode import unidecode

from .data.certificates_data import (
    certificate_countries_as_list,
    certificates_as_df,
    certificates_as_dict,
    certificates_as_list,
)
from .utils.data import Data
from .utils.extra import random_user_agent


def get_certificates(country=None):
    """
    This function retrieves all the data stored in `certificates.csv` file, which previously was retrieved from
    Investing.com. Since the resulting object is a matrix of data, the certificate's data is properly structured
    in rows and columns, where columns are the certificate data attribute names. Additionally, country
    filtering can be specified, which will make this function return not all the stored certificates, but just
    the data of the certificates from the introduced country.

    Args:
        country (:obj:`str`, optional): name of the country to retrieve all its available certificates from.

    Returns:
        :obj:`pandas.DataFrame` - certificates_df:
            The resulting :obj:`pandas.DataFrame` contains all the certificate's data from the introduced country if specified,
            or from every country if None was specified, as indexed in Investing.com from the information previously
            retrieved by investpy and stored on a csv file.

            So on, the resulting :obj:`pandas.DataFrame` will look like::

                country | name | full_name | symbol | issuer | isin | asset_class | underlying
                --------|------|-----------|--------|--------|------|-------------|------------
                xxxxxxx | xxxx | xxxxxxxxx | xxxxxx | xxxxxx | xxxx | xxxxxxxxxxx | xxxxxxxxxx

    Raises:
        ValueError: raised whenever any of the introduced arguments is not valid.
        FileNotFoundError: raised if `certificates.csv` file was not found.
        IOError: raised when `certificates.csv` file is missing or empty.

    """

    return certificates_as_df(country)


def get_certificates_list(country=None):
    """
    This function retrieves all the available certificates indexed on Investing.com, already stored on `certificates.csv`.
    This function also allows the users to specify which country do they want to retrieve data from or if they
    want to retrieve it from every listed country; so on, a listing of certificates will be returned. This function
    helps the user to get to know which certificates are available on Investing.com.

    Args:
        country (:obj:`str`, optional): name of the country to retrieve all its available certificates from.

    Returns:
        :obj:`list` - certificates_list:
            The resulting :obj:`list` contains the retrieved data from the `certificates.csv` file, which is
            a listing of the names of the certificates listed on Investing.com, which is the input for data
            retrieval functions as the name of the certificate to retrieve data from needs to be specified.

            In case the listing was successfully retrieved, the :obj:`list` will look like::

                certificates_list = ['SOCIETE GENERALE CAC 40 X10 31DEC99', 'SG ZT CAC 40 x7 Short 31Dec99', ...]

    Raises:
        ValueError: raised whenever any of the introduced arguments is not valid.
        FileNotFoundError: raised if `certificates.csv` file was not found.
        IOError: raised when `certificates.csv` file is missing or empty.

    """

    return certificates_as_list(country)


def get_certificates_dict(country=None, columns=None, as_json=False):
    """
    This function retrieves all the available certificates indexed on Investing.com, stored on `certificates.csv`.
    This function also allows the user to specify which country do they want to retrieve data from, or from every
    listed country; the columns which the user wants to be included on the resulting :obj:`dict`; and the output
    of the function will either be a :obj:`dict` or a :obj:`json`.

    Args:
        country (:obj:`str`, optional): name of the country to retrieve all its available certificates from.
        columns (:obj:`list`, optional):
            names of the columns of the certificate data to retrieve <country, name, full_name, symbol, issuer, isin, asset_class, underlying>
        as_json (:obj:`bool`, optional):
            value to determine the format of the output data which can either be a :obj:`dict` or a :obj:`json`.

    Returns:
        :obj:`dict` or :obj:`json` - certificates_dict:
            The resulting :obj:`dict` contains the retrieved data if found, if not, the corresponding fields are
            filled with `None` values.

            In case the information was successfully retrieved, the :obj:`dict` will look like::

                certificates_dict = {
                    "country": "france",
                    "name": "SOCIETE GENERALE CAC 40 X10 31DEC99",
                    "full_name": "SOCIETE GENERALE EFFEKTEN GMBH ZT CAC 40 X10 LEVERAGE 31DEC99",
                    "symbol": "FR0011214527",
                    "issuer": "Societe Generale Effekten GMBH",
                    "isin": "FR0011214527",
                    "asset_class": "index",
                    "underlying": "CAC 40 Leverage x10 NR"
                }

    Raises:
        ValueError: raised whenever any of the introduced arguments is not valid.
        FileNotFoundError: raised if `certificates.csv` file was not found.
        IOError: raised when `certificates.csv` file is missing or empty.

    """

    return certificates_as_dict(country=country, columns=columns, as_json=as_json)


def get_certificate_countries():
    """
    This function retrieves all the available countries to retrieve certificates from, as the listed countries
    are the ones indexed on Investing.com. The purpose of this function is to list the countries which
    have available certificates according to Investing.com data, since the country parameter is needed when
    retrieving data from any certificate available.

    Returns:
        :obj:`list` - countries:
            The resulting :obj:`list` contains all the countries listed on Investing.com with available certificates
            to retrieve data from.

            In the case that the file reading of `certificate_countries.csv` which contains the names of the available
            countries with certificates was successfully completed, the resulting :obj:`list` will look like::

                countries = ['france', 'germany', 'italy', 'netherlands', 'sweden']

    Raises:
        FileNotFoundError: raised when `certificate_countries.csv` file was not found.

    """

    return certificate_countries_as_list()


def get_certificate_recent_data(
    certificate, country, as_json=False, order="ascending", interval="Daily"
):
    """
    This function retrieves recent historical data from the introduced certificate from Investing.com. So on, the recent data
    of the introduced certificate from the specified country will be retrieved and returned as a :obj:`pandas.DataFrame` if
    the parameters are valid and the request to Investing.com succeeds. Note that additionally some optional parameters
    can be specified: as_json and order, which let the user decide if the data is going to be returned as a
    :obj:`json` or not, and if the historical data is going to be ordered ascending or descending (where the index is the
    date), respectively.

    Args:
        certificate (:obj:`str`): name of the certificate to retrieve recent data from.
        country (:obj:`str`): name of the country from where the certificate is.
        as_json (:obj:`bool`, optional):
            to determine the format of the output data, either a :obj:`pandas.DataFrame` if False and a :obj:`json` if True.
        order (:obj:`str`, optional): to define the order of the retrieved data which can either be ascending or descending.
        interval (:obj:`str`, optional):
            value to define the historical data interval to retrieve, by default `Daily`, but it can also be `Weekly` or `Monthly`.

    Returns:
        :obj:`pandas.DataFrame` or :obj:`json`:
            The function returns either a :obj:`pandas.DataFrame` or a :obj:`json` file containing the retrieved recent
            data from the specified certificate via argument. The dataset contains the OHLC values of the certificate.

            The returned data is case we use default arguments will look like::

                Date || Open | High | Low | Close
                -----||------|------|-----|-------
                xxxx || xxxx | xxxx | xxx | xxxxx

            but if we define `as_json=True`, then the output will be::

                {
                    name: name,
                    recent: [
                        {
                            date: dd/mm/yyyy,
                            open: x,
                            high: x,
                            low: x,
                            close: x
                        },
                        ...
                    ]
                }

    Raises:
        ValueError: raised if there was an argument error.
        IOError: raised if certificates object/file was not found or unable to retrieve.
        RuntimeError: raised if the introduced certificate does not match any of the indexed ones.
        ConnectionError: raised if GET requests does not return 200 status code.
        IndexError: raised if certificate information was unavailable or not found.

    Examples:
        >>> data = investpy.get_certificate_recent_data(certificate='BNP Gold 31Dec99', country='france')
        >>> data.head()
                     Open   High     Low   Close
        Date
        2020-07-09  146.4  146.8  145.95  145.95
        2020-07-10  146.2  146.2  145.55  145.55
        2020-07-13  145.6  145.6  145.45  145.45
        2020-07-14  145.4  145.4  145.25  145.25
        2020-07-15  144.9  145.1  144.70  144.95

    """

    if not certificate:
        raise ValueError(
            "ERR#0100: certificate param is mandatory and should be a str."
        )

    if not isinstance(certificate, str):
        raise ValueError(
            "ERR#0100: certificate param is mandatory and should be a str."
        )

    if country is None:
        raise ValueError("ERR#0039: country can not be None, it should be a str.")

    if country is not None and not isinstance(country, str):
        raise ValueError("ERR#0025: specified country value not valid.")

    if not isinstance(as_json, bool):
        raise ValueError(
            "ERR#0002: as_json argument can just be True or False, bool type."
        )

    if order not in ["ascending", "asc", "descending", "desc"]:
        raise ValueError(
            "ERR#0003: order argument can just be ascending (asc) or descending (desc),"
            " str type."
        )

    if not interval:
        raise ValueError(
            "ERR#0073: interval value should be a str type and it can just be either"
            " 'Daily', 'Weekly' or 'Monthly'."
        )

    if not isinstance(interval, str):
        raise ValueError(
            "ERR#0073: interval value should be a str type and it can just be either"
            " 'Daily', 'Weekly' or 'Monthly'."
        )

    interval = interval.lower()

    if interval not in ["daily", "weekly", "monthly"]:
        raise ValueError(
            "ERR#0073: interval value should be a str type and it can just be either"
            " 'Daily', 'Weekly' or 'Monthly'."
        )

    resource_package = "investpy"
    resource_path = "/".join(("resources", "certificates.csv"))
    if pkg_resources.resource_exists(resource_package, resource_path):
        certificates = pd.read_csv(
            pkg_resources.resource_filename(resource_package, resource_path),
            keep_default_na=False,
        )
    else:
        raise FileNotFoundError("ERR#0096: certificates file not found or errored.")

    if certificates is None:
        raise IOError("ERR#0097: certificates not found or unable to retrieve.")

    country = unidecode(country.strip().lower())

    if country not in get_certificate_countries():
        raise RuntimeError(
            "ERR#0034: country "
            + country.lower()
            + " not found, check if it is correct."
        )

    certificates = certificates[certificates["country"] == country]

    certificate = unidecode(certificate.strip().lower())

    if certificate not in list(certificates["name"].apply(unidecode).str.lower()):
        raise RuntimeError(
            "ERR#0101: certificate "
            + certificate
            + " not found, check if it is correct."
        )

    symbol = certificates.loc[
        (certificates["name"].apply(unidecode).str.lower() == certificate).idxmax(),
        "symbol",
    ]
    id_ = certificates.loc[
        (certificates["name"].apply(unidecode).str.lower() == certificate).idxmax(),
        "id",
    ]
    name = certificates.loc[
        (certificates["name"].apply(unidecode).str.lower() == certificate).idxmax(),
        "name",
    ]

    header = symbol + " Historical Data"

    params = {
        "curr_id": id_,
        "smlID": str(randint(1000000, 99999999)),
        "header": header,
        "interval_sec": interval.capitalize(),
        "sort_col": "date",
        "sort_ord": "DESC",
        "action": "historical_data",
    }

    head = {
        "User-Agent": random_user_agent(),
        "X-Requested-With": "XMLHttpRequest",
        "Accept": "text/html",
        "Accept-Encoding": "gzip, deflate",
        "Connection": "keep-alive",
    }

    url = "https://www.investing.com/instruments/HistoricalDataAjax"

    req = requests.post(url, headers=head, data=params)

    if req.status_code != 200:
        raise ConnectionError(
            "ERR#0015: error " + str(req.status_code) + ", try again later."
        )

    root_ = fromstring(req.text)
    path_ = root_.xpath(".//table[@id='curr_table']/tbody/tr")

    result = list()

    if path_:
        for elements_ in path_:
            if elements_.xpath(".//td")[0].text_content() == "No results found":
                raise IndexError(
                    "ERR#0102: certificate information unavailable or not found."
                )

            info = []

            for nested_ in elements_.xpath(".//td"):
                info.append(nested_.get("data-real-value"))

            certificate_date = datetime.strptime(
                str(
                    datetime.fromtimestamp(int(info[0]), tz=pytz.timezone("GMT")).date()
                ),
                "%Y-%m-%d",
            )

            certificate_close = float(info[1].replace(",", ""))
            certificate_open = float(info[2].replace(",", ""))
            certificate_high = float(info[3].replace(",", ""))
            certificate_low = float(info[4].replace(",", ""))

            result.insert(
                len(result),
                Data(
                    certificate_date,
                    certificate_open,
                    certificate_high,
                    certificate_low,
                    certificate_close,
                    None,
                    None,
                    None,
                ),
            )

        if order in ["ascending", "asc"]:
            result = result[::-1]
        elif order in ["descending", "desc"]:
            result = result

        if as_json is True:
            json_ = {
                "name": name,
                "recent": [value.certificate_as_json() for value in result],
            }

            return json.dumps(json_, sort_keys=False)
        elif as_json is False:
            df = pd.DataFrame.from_records(
                [value.certificate_to_dict() for value in result]
            )
            df.set_index("Date", inplace=True)

            return df
    else:
        raise RuntimeError("ERR#0004: data retrieval error while scraping.")


def get_certificate_historical_data(
    certificate,
    country,
    from_date,
    to_date,
    as_json=False,
    order="ascending",
    interval="Daily",
):
    """
    This function retrieves historical data from the introduced certificate from Investing.com. So on, the historical data
    of the introduced certificate from the specified country in the specified date range will be retrieved and returned as
    a :obj:`pandas.DataFrame` if the parameters are valid and the request to Investing.com succeeds. Note that additionally
    some optional parameters can be specified: as_json and order, which let the user decide if the data is going to
    be returned as a :obj:`json` or not, and if the historical data is going to be ordered ascending or descending (where the
    index is the date), respectively.

    Args:
        certificate (:obj:`str`): name of the certificate to retrieve historical data from.
        country (:obj:`str`): name of the country from where the certificate is.
        from_date (:obj:`str`): date formatted as `dd/mm/yyyy`, since when data is going to be retrieved.
        to_date (:obj:`str`): date formatted as `dd/mm/yyyy`, until when data is going to be retrieved.
        as_json (:obj:`bool`, optional):
            to determine the format of the output data, either a :obj:`pandas.DataFrame` if False and a :obj:`json` if True.
        order (:obj:`str`, optional): to define the order of the retrieved data which can either be ascending or descending.
        interval (:obj:`str`, optional):
            value to define the historical data interval to retrieve, by default `Daily`, but it can also be `Weekly` or `Monthly`.

    Returns:
        :obj:`pandas.DataFrame` or :obj:`json`:
            The function can return either a :obj:`pandas.DataFrame` or a :obj:`json` object, containing the retrieved
            historical data of the specified certificate from the specified country. So on, the resulting dataframe contains the
            OHLC values for the selected certificate on market days.

            The returned data is case we use default arguments will look like::

                Date || Open | High | Low | Close
                -----||------|------|-----|-------
                xxxx || xxxx | xxxx | xxx | xxxxx

            but if we define `as_json=True`, then the output will be::

                {
                    name: name,
                    historical: [
                        {
                            date: 'dd/mm/yyyy',
                            open: x,
                            high: x,
                            low: x,
                            close: x
                        },
                        ...
                    ]
                }

    Raises:
        ValueError: raised whenever any of the introduced arguments is not valid or errored.
        IOError: raised if certificates object/file was not found or unable to retrieve.
        RuntimeError: raised if the introduced certificate/country was not found or did not match any of the existing ones.
        ConnectionError: raised if connection to Investing.com could not be established.
        IndexError: raised if certificate historical data was unavailable or not found in Investing.com.

    Examples:
        >>> data = investpy.get_certificate_historical_data(certificate='BNP Gold 31Dec99', country='france', from_date='01/01/2010', to_date='01/01/2019')
        >>> data.head()
                     Open   High    Low  Close
        Date
        2010-01-04  77.15  77.15  77.15  77.15
        2010-01-05  77.40  77.45  77.15  77.45
        2010-01-06  78.40  78.40  78.40  78.40
        2010-01-07  78.40  78.45  78.35  78.35
        2010-01-08  77.95  78.10  77.95  78.10

    """

    if not certificate:
        raise ValueError(
            "ERR#0100: certificate param is mandatory and should be a str."
        )

    if not isinstance(certificate, str):
        raise ValueError(
            "ERR#0100: certificate param is mandatory and should be a str."
        )

    if country is None:
        raise ValueError("ERR#0039: country can not be None, it should be a str.")

    if country is not None and not isinstance(country, str):
        raise ValueError("ERR#0025: specified country value not valid.")

    if not isinstance(as_json, bool):
        raise ValueError(
            "ERR#0002: as_json argument can just be True or False, bool type."
        )

    if order not in ["ascending", "asc", "descending", "desc"]:
        raise ValueError(
            "ERR#0003: order argument can just be ascending (asc) or descending (desc),"
            " str type."
        )

    if not interval:
        raise ValueError(
            "ERR#0073: interval value should be a str type and it can just be either"
            " 'Daily', 'Weekly' or 'Monthly'."
        )

    if not isinstance(interval, str):
        raise ValueError(
            "ERR#0073: interval value should be a str type and it can just be either"
            " 'Daily', 'Weekly' or 'Monthly'."
        )

    interval = interval.lower()

    if interval not in ["daily", "weekly", "monthly"]:
        raise ValueError(
            "ERR#0073: interval value should be a str type and it can just be either"
            " 'Daily', 'Weekly' or 'Monthly'."
        )

    try:
        datetime.strptime(from_date, "%d/%m/%Y")
    except ValueError:
        raise ValueError(
            "ERR#0011: incorrect from_date date format, it should be 'dd/mm/yyyy'."
        )

    try:
        datetime.strptime(to_date, "%d/%m/%Y")
    except ValueError:
        raise ValueError(
            "ERR#0012: incorrect to_date format, it should be 'dd/mm/yyyy'."
        )

    start_date = datetime.strptime(from_date, "%d/%m/%Y")
    end_date = datetime.strptime(to_date, "%d/%m/%Y")

    if start_date >= end_date:
        raise ValueError(
            "ERR#0032: to_date should be greater than from_date, both formatted as"
            " 'dd/mm/yyyy'."
        )

    date_interval = {
        "intervals": [],
    }

    flag = True

    while flag is True:
        diff = end_date.year - start_date.year

        if diff > 19:
            obj = {
                "start": start_date.strftime("%m/%d/%Y"),
                "end": start_date.replace(year=start_date.year + 19).strftime(
                    "%m/%d/%Y"
                ),
            }

            date_interval["intervals"].append(obj)

            start_date = start_date.replace(year=start_date.year + 19) + timedelta(
                days=1
            )
        else:
            obj = {
                "start": start_date.strftime("%m/%d/%Y"),
                "end": end_date.strftime("%m/%d/%Y"),
            }

            date_interval["intervals"].append(obj)

            flag = False

    interval_limit = len(date_interval["intervals"])
    interval_counter = 0

    data_flag = False

    resource_package = "investpy"
    resource_path = "/".join(("resources", "certificates.csv"))
    if pkg_resources.resource_exists(resource_package, resource_path):
        certificates = pd.read_csv(
            pkg_resources.resource_filename(resource_package, resource_path),
            keep_default_na=False,
        )
    else:
        raise FileNotFoundError("ERR#0096: certificates file not found or errored.")

    if certificates is None:
        raise IOError("ERR#0097: certificates not found or unable to retrieve.")

    country = unidecode(country.strip().lower())

    if country not in get_certificate_countries():
        raise RuntimeError(
            "ERR#0034: country "
            + country.lower()
            + " not found, check if it is correct."
        )

    certificates = certificates[certificates["country"] == country]

    certificate = unidecode(certificate.strip().lower())

    if certificate not in list(certificates["name"].apply(unidecode).str.lower()):
        raise RuntimeError(
            "ERR#0101: certificate "
            + certificate
            + " not found, check if it is correct."
        )

    symbol = certificates.loc[
        (certificates["name"].apply(unidecode).str.lower() == certificate).idxmax(),
        "symbol",
    ]
    id_ = certificates.loc[
        (certificates["name"].apply(unidecode).str.lower() == certificate).idxmax(),
        "id",
    ]
    name = certificates.loc[
        (certificates["name"].apply(unidecode).str.lower() == certificate).idxmax(),
        "name",
    ]

    header = symbol + " Historical Data"

    final = list()

    for index in range(len(date_interval["intervals"])):
        interval_counter += 1

        params = {
            "curr_id": id_,
            "smlID": str(randint(1000000, 99999999)),
            "header": header,
            "st_date": date_interval["intervals"][index]["start"],
            "end_date": date_interval["intervals"][index]["end"],
            "interval_sec": interval.capitalize(),
            "sort_col": "date",
            "sort_ord": "DESC",
            "action": "historical_data",
        }

        head = {
            "User-Agent": random_user_agent(),
            "X-Requested-With": "XMLHttpRequest",
            "Accept": "text/html",
            "Accept-Encoding": "gzip, deflate",
            "Connection": "keep-alive",
        }

        url = "https://www.investing.com/instruments/HistoricalDataAjax"

        req = requests.post(url, headers=head, data=params)

        if req.status_code != 200:
            raise ConnectionError(
                "ERR#0015: error " + str(req.status_code) + ", try again later."
            )

        if not req.text:
            continue

        root_ = fromstring(req.text)
        path_ = root_.xpath(".//table[@id='curr_table']/tbody/tr")

        result = list()

        if path_:
            for elements_ in path_:
                if elements_.xpath(".//td")[0].text_content() == "No results found":
                    if interval_counter < interval_limit:
                        data_flag = False
                    else:
                        raise IndexError(
                            "ERR#0102: certificate information unavailable or not"
                            " found."
                        )
                else:
                    data_flag = True

                info = []

                for nested_ in elements_.xpath(".//td"):
                    info.append(nested_.get("data-real-value"))

                if data_flag is True:
                    certificate_date = datetime.strptime(
                        str(
                            datetime.fromtimestamp(
                                int(info[0]), tz=pytz.timezone("GMT")
                            ).date()
                        ),
                        "%Y-%m-%d",
                    )

                    certificate_close = float(info[1].replace(",", ""))
                    certificate_open = float(info[2].replace(",", ""))
                    certificate_high = float(info[3].replace(",", ""))
                    certificate_low = float(info[4].replace(",", ""))

                    result.insert(
                        len(result),
                        Data(
                            certificate_date,
                            certificate_open,
                            certificate_high,
                            certificate_low,
                            certificate_close,
                            None,
                            None,
                            None,
                        ),
                    )

            if data_flag is True:
                if order in ["ascending", "asc"]:
                    result = result[::-1]
                elif order in ["descending", "desc"]:
                    result = result

                if as_json is True:
                    json_list = [value.certificate_as_json() for value in result]

                    final.append(json_list)
                elif as_json is False:
                    df = pd.DataFrame.from_records(
                        [value.certificate_to_dict() for value in result]
                    )
                    df.set_index("Date", inplace=True)

                    final.append(df)
        else:
            raise RuntimeError("ERR#0004: data retrieval error while scraping.")

    if order in ["descending", "desc"]:
        final.reverse()

    if as_json is True:
        json_ = {
            "name": name,
            "historical": [value for json_list in final for value in json_list],
        }
        return json.dumps(json_, sort_keys=False)
    elif as_json is False:
        return pd.concat(final)


def get_certificate_information(certificate, country, as_json=False):
    """
    This function retrieves fundamental financial information from the specified certificate. The retrieved
    information from the certificate can be valuable as it is additional information that can be used combined
    with OHLC values, so to determine financial insights from the company which holds the specified certificate.

    Args:
        certificate (:obj:`str`): name of the certificate to retrieve information from
        country (:obj:`country`): name of the country from where the certificate is from.
        as_json (:obj:`bool`, optional):
            optional argument to determine the format of the output data (:obj:`dict` or :obj:`json`).

    Returns:
        :obj:`pandas.DataFrame` or :obj:`dict`- certificate_information:
            The resulting :obj:`pandas.DataFrame` contains the information fields retrieved from Investing.com
            from the specified certificate; it can also be returned as a :obj:`dict`, if argument `as_json=True`.

            If any of the information fields could not be retrieved, that field/s will be filled with
            None values. If the retrieval process succeeded, the resulting :obj:`dict` will look like::

                certificate_information = {
                    "Certificate Name": "XXXX",
                    "Certificate Country": "XXXX",
                    "Prev. Close": X.Y,
                    "Todays Range": "X.Y - X.Y",
                    "Leverage": "X:Y",
                    "Open": X.Y,
                    "52 wk Range": "X.Y - X.Y",
                    "Strike Price": "XXXX",
                    "Volume": X.Y,
                    "Issue Date": "XXXX",
                    "Issue Amount": "XXXX",
                    "Average Vol. (3m)": X.Y,
                    "Maturity Date": "dd/mm/yyyy",
                    "1-Year Change": "X.Y%",
                    "Asset Class": "XXXX"
                }

    """

    if not certificate:
        raise ValueError(
            "ERR#0100: certificate param is mandatory and should be a str."
        )

    if not isinstance(certificate, str):
        raise ValueError(
            "ERR#0100: certificate param is mandatory and should be a str."
        )

    if country is None:
        raise ValueError("ERR#0039: country can not be None, it should be a str.")

    if country is not None and not isinstance(country, str):
        raise ValueError("ERR#0025: specified country value not valid.")

    if not isinstance(as_json, bool):
        raise ValueError(
            "ERR#0002: as_json argument can just be True or False, bool type."
        )

    resource_package = "investpy"
    resource_path = "/".join(("resources", "certificates.csv"))
    if pkg_resources.resource_exists(resource_package, resource_path):
        certificates = pd.read_csv(
            pkg_resources.resource_filename(resource_package, resource_path),
            keep_default_na=False,
        )
    else:
        raise FileNotFoundError("ERR#0096: certificates file not found or errored.")

    if certificates is None:
        raise IOError("ERR#0097: certificates not found or unable to retrieve.")

    country = unidecode(country.strip().lower())

    if country not in get_certificate_countries():
        raise RuntimeError(
            "ERR#0034: country "
            + country.lower()
            + " not found, check if it is correct."
        )

    certificates = certificates[certificates["country"] == country]

    certificate = unidecode(certificate.strip().lower())

    if certificate not in list(certificates["name"].apply(unidecode).str.lower()):
        raise RuntimeError(
            "ERR#0101: certificate "
            + certificate
            + " not found, check if it is correct."
        )

    tag = certificates.loc[
        (certificates["name"].apply(unidecode).str.lower() == certificate).idxmax(),
        "tag",
    ]
    name = certificates.loc[
        (certificates["name"].apply(unidecode).str.lower() == certificate).idxmax(),
        "name",
    ]

    url = "https://www.investing.com/certificates/" + tag

    head = {
        "User-Agent": random_user_agent(),
        "X-Requested-With": "XMLHttpRequest",
        "Accept": "text/html",
        "Accept-Encoding": "gzip, deflate",
        "Connection": "keep-alive",
    }

    req = requests.get(url, headers=head)

    if req.status_code != 200:
        raise ConnectionError(
            "ERR#0015: error " + str(req.status_code) + ", try again later."
        )

    root_ = fromstring(req.text)
    path_ = root_.xpath("//div[contains(@class, 'overviewDataTable')]/div")

    result = pd.DataFrame(
        columns=[
            "Certificate Name",
            "Certificate Country",
            "Prev. Close",
            "Todays Range",
            "Leverage",
            "Open",
            "52 wk Range",
            "Strike Price",
            "Volume",
            "Issue Date",
            "Issue Amount",
            "Average Vol. (3m)",
            "Maturity Date",
            "1-Year Change",
            "Asset Class",
        ]
    )
    result.at[0, "Certificate Name"] = name
    result.at[0, "Certificate Country"] = country

    if path_:
        for elements_ in path_:
            element = elements_.xpath(".//span[@class='float_lang_base_1']")[0]
            title_ = element.text_content()
            if title_ == "Day's Range":
                title_ = "Todays Range"
            if title_ in result.columns.tolist():
                try:
                    result.at[0, title_] = float(
                        element.getnext().text_content().replace(",", "")
                    )
                    continue
                except:
                    pass
                try:
                    text = element.getnext().text_content().strip()
                    result.at[0, title_] = datetime.strptime(
                        text, "%b %d, %Y"
                    ).strftime("%d/%m/%Y")
                    continue
                except:
                    pass
                try:
                    value = element.getnext().text_content().strip()
                    if value.__contains__("K"):
                        value = float(value.replace("K", "").replace(",", "")) * 1e3
                    elif value.__contains__("M"):
                        value = float(value.replace("M", "").replace(",", "")) * 1e6
                    elif value.__contains__("B"):
                        value = float(value.replace("B", "").replace(",", "")) * 1e9
                    elif value.__contains__("T"):
                        value = float(value.replace("T", "").replace(",", "")) * 1e12
                    result.at[0, title_] = value
                    continue
                except:
                    pass

        result.replace({"N/A": None}, inplace=True)

        if as_json is True:
            json_ = result.iloc[0].to_dict()
            return json_
        elif as_json is False:
            return result
    else:
        raise RuntimeError("ERR#0004: data retrieval error while scraping.")


def get_certificates_overview(country, as_json=False, n_results=100):
    """
    This function retrieves an overview containing all the real time data available for the main certificates
    from a country, such as the names, symbols, current value, etc. as indexed in Investing.com. So on, the main
    usage of this function is to get an overview on the main certificates from a country, so to get a general view.
    Note that since this function is retrieving a lot of information at once, by default just the overview of
    the Top 100 certificates is being retrieved, but an additional parameter called n_results can be specified so to
    retrieve N results.

    Args:
        country (:obj:`str`): name of the country to retrieve the certificates overview from.
        as_json (:obj:`bool`, optional):
            optional argument to determine the format of the output data (:obj:`pandas.DataFrame` or :obj:`json`).
        n_results (:obj:`int`, optional): number of results to be displayed on the overview table (0-1000).

    Returns:
        :obj:`pandas.DataFrame` - certificates_overview:
            The resulting :obj:`pandas.DataFrame` contains all the data available in Investing.com of the main certificates
            from a country in order to get an overview of it.

            If the retrieval process succeeded, the resulting :obj:`pandas.DataFrame` should look like::

                country | name | symbol | last | change_percentage | turnover
                --------|------|--------|------|-------------------|----------
                xxxxxxx | xxxx | xxxxxx | xxxx | xxxxxxxxxxxxxxxxx | xxxxxxxx

    Raises:
        ValueError: raised if any of the introduced arguments is not valid or errored.
        FileNotFoundError: raised when `certificates.csv` file is missing.
        IOError: raised if data could not be retrieved due to file error.
        RuntimeError:
            raised either if the introduced country does not match any of the listed ones or if no overview results could be
            retrieved from Investing.com.
        ConnectionError: raised if GET requests does not return 200 status code.

    """

    if country is None:
        raise ValueError("ERR#0039: country can not be None, it should be a str.")

    if country is not None and not isinstance(country, str):
        raise ValueError("ERR#0025: specified country value not valid.")

    if not isinstance(as_json, bool):
        raise ValueError(
            "ERR#0002: as_json argument can just be True or False, bool type."
        )

    if not isinstance(n_results, int):
        raise ValueError(
            "ERR#0089: n_results argument should be an integer between 1 and 1000."
        )

    if 1 > n_results or n_results > 1000:
        raise ValueError(
            "ERR#0089: n_results argument should be an integer between 1 and 1000."
        )

    resource_package = "investpy"
    resource_path = "/".join(("resources", "certificates.csv"))
    if pkg_resources.resource_exists(resource_package, resource_path):
        certificates = pd.read_csv(
            pkg_resources.resource_filename(resource_package, resource_path),
            keep_default_na=False,
        )
    else:
        raise FileNotFoundError("ERR#0096: certificates file not found or errored.")

    if certificates is None:
        raise IOError("ERR#0097: certificates not found or unable to retrieve.")

    country = unidecode(country.strip().lower())

    if country not in get_certificate_countries():
        raise RuntimeError(
            "ERR#0034: country "
            + country.lower()
            + " not found, check if it is correct."
        )

    certificates = certificates[certificates["country"] == country]

    head = {
        "User-Agent": random_user_agent(),
        "X-Requested-With": "XMLHttpRequest",
        "Accept": "text/html",
        "Accept-Encoding": "gzip, deflate",
        "Connection": "keep-alive",
    }

    url = (
        "https://www.investing.com/certificates/"
        + country.replace(" ", "-")
        + "-certificates"
    )

    req = requests.get(url, headers=head)

    if req.status_code != 200:
        raise ConnectionError(
            "ERR#0015: error " + str(req.status_code) + ", try again later."
        )

    root_ = fromstring(req.text)
    table = root_.xpath(".//table[@id='certificates']/tbody/tr")

    results = list()

    if len(table) > 0:
        for row in table[:n_results]:
            id_ = row.get("id").replace("pair_", "")
            country_check = (
                row.xpath(".//td[@class='flag']/span")[0].get("title").lower()
            )

            name = (
                row.xpath(".//td[contains(@class, 'elp')]/a")[0].text_content().strip()
            )

            pid = "pid-" + id_

            symbol = (
                row.xpath(".//td[contains(@class, 'symbol')]")[0].get("title").strip()
            )

            last = row.xpath(".//td[@class='" + pid + "-last']")[0].text_content()

            pcp = row.xpath(".//td[contains(@class, '" + pid + "-pcp')]")[
                0
            ].text_content()
            turnover = row.xpath(".//td[contains(@class, '" + pid + "-turnover')]")[
                0
            ].text_content()

            if turnover != "0":
                if turnover.__contains__("K"):
                    turnover = float(turnover.replace("K", "").replace(",", "")) * 1e3
                elif turnover.__contains__("M"):
                    turnover = float(turnover.replace("M", "").replace(",", "")) * 1e6
                elif turnover.__contains__("B"):
                    turnover = float(turnover.replace("B", "").replace(",", "")) * 1e9
                else:
                    turnover = float(turnover.replace(",", ""))

            data = {
                "country": country_check,
                "name": name,
                "symbol": symbol,
                "last": float(last.replace(",", "")),
                "change_percentage": pcp,
                "turnover": int(turnover),
            }

            results.append(data)
    else:
        raise RuntimeError(
            "ERR#0092: no data found while retrieving the overview from Investing.com"
        )

    df = pd.DataFrame(results)

    if as_json:
        return json.loads(df.to_json(orient="records"))
    else:
        return df


def search_certificates(by, value):
    """
    This function searches certificates by the introduced value for the specified field. This means that this function
    is going to search if there is a value that matches the introduced one for the specified field which is the
    `certificates.csv` column name to search in. Available fields to search certificates are `country`, `name`,
    `full_name`, `symbol`, `issuer`, `isin`, `asset_class`, `underlying`.

    Args:
        by (:obj:`str`):
            name of the field to search for, which is the column name which can be: country, name, full_name, symbol,
            issuer, isin, asset_class or underlying.
        value (:obj:`str`): value of the field to search for, which is the value that is going to be searched.

    Returns:
        :obj:`pandas.DataFrame` - search_result:
            The resulting :obj:`pandas.DataFrame` contains the search results from the given query, which is
            any match of the specified value in the specified field. If there are no results for the given query,
            an error will be raised, but otherwise the resulting :obj:`pandas.DataFrame` will contain all the
            available certificates that match the introduced query.

    Raises:
        ValueError: raised if any of the introduced parameters is not valid or errored.
        FileNotFoundError: raised if `certificates.csv` file is missing.
        IOError: raised if data could not be retrieved due to file error.
        RuntimeError: raised if no results were found for the introduced value in the introduced field.

    """

    if not by:
        raise ValueError(
            "ERR#0006: the introduced field to search is mandatory and should be a str."
        )

    if not isinstance(by, str):
        raise ValueError(
            "ERR#0006: the introduced field to search is mandatory and should be a str."
        )

    if not value:
        raise ValueError(
            "ERR#0017: the introduced value to search is mandatory and should be a str."
        )

    if not isinstance(value, str):
        raise ValueError(
            "ERR#0017: the introduced value to search is mandatory and should be a str."
        )

    resource_package = "investpy"
    resource_path = "/".join(("resources", "certificates.csv"))
    if pkg_resources.resource_exists(resource_package, resource_path):
        certificates = pd.read_csv(
            pkg_resources.resource_filename(resource_package, resource_path),
            keep_default_na=False,
        )
    else:
        raise FileNotFoundError("ERR#0096: certificates file not found or errored.")

    if certificates is None:
        raise IOError("ERR#0097: certificates not found or unable to retrieve.")

    certificates.drop(columns=["tag", "id"], inplace=True)

    available_search_fields = certificates.columns.tolist()

    if isinstance(by, str) and by not in available_search_fields:
        raise ValueError(
            "ERR#0026: the introduced field to search can either just be "
            + " or ".join(available_search_fields)
        )

    certificates["matches"] = certificates[by].str.contains(value, case=False)

    search_result = certificates.loc[certificates["matches"] == True].copy()

    if len(search_result) == 0:
        raise RuntimeError(
            "ERR#0043: no results were found for the introduced " + str(by) + "."
        )

    search_result.drop(columns=["matches"], inplace=True)
    search_result.reset_index(drop=True, inplace=True)

    return search_result



# Copyright 2018-2021 Alvaro Bartolome, alvarobartt @ GitHub
# See LICENSE for details.

import json

import pandas as pd
import pkg_resources
from unidecode import unidecode

from ..utils import constant as cst


def currency_crosses_as_df(base=None, second=None):
    """
    This function retrieves all the available currency crosses from Investing.com and returns them as a
    :obj:`pandas.DataFrame`, which contains not just the currency crosses names, but all the fields contained on
    the currency_crosses file. Note that the filtering params are both base and second, which mean the base and the
    second currency of the currency cross, for example, in the currency cross `EUR/USD` the base currency is EUR and
    the second currency is USD. These are optional parameters, so specifying one of them means that all the currency
    crosses where the introduced currency is either base or second will be returned; if both are specified,
    just the introduced currency cross will be returned if it exists. All the available currency crosses can be found
    at: https://www.investing.com/currencies/

    Args:
        base (:obj:`str`, optional):
            symbol of the base currency of the currency cross, this will return a :obj:`pandas.DataFrame` containing
            all the currency crosses where the base currency matches the introduced one.
        second (:obj:`str`):
            symbol of the second currency of the currency cross, this will return a :obj:`pandas.DataFrame` containing
            all the currency crosses where the second currency matches the introduced one.

    Returns:
        :obj:`pandas.DataFrame` - currency_crosses_df:
            The resulting :obj:`pandas.DataFrame` contains all the currency crosses basic information retrieved from
            Investing.com.

            In case the information was successfully retrieved, the resulting :obj:`pandas.DataFrame` will look like::

                name | full_name | base | second | base_name | second_name
                -----|-----------|------|--------|-----------|-------------
                xxxx | xxxxxxxxx | xxxx | xxxxxx | xxxxxxxxx | xxxxxxxxxxx

    Raises:
        ValueError: raised if any of the introduced arguments is not valid or errored.
        FileNotFoundError: raised if `currency_crosses.csv` file was not found.
        IOError: raised if currency crosses retrieval failed, both for missing file or empty file.

    """

    if base is not None and not isinstance(base, str):
        raise ValueError("ERR#0049: specified base currency value is not valid.")

    if second is not None and not isinstance(second, str):
        raise ValueError("ERR#0051: specified second currency value is not valid.")

    resource_package = "investpy"
    resource_path = "/".join(("resources", "currency_crosses.csv"))
    if pkg_resources.resource_exists(resource_package, resource_path):
        currency_crosses = pd.read_csv(
            pkg_resources.resource_filename(resource_package, resource_path),
            keep_default_na=False,
        )
    else:
        raise FileNotFoundError("ERR#0060: currency_crosses file not found or errored.")

    if currency_crosses is None:
        raise IOError("ERR#0050: currency_crosses not found or unable to retrieve.")

    available_currencies = available_currencies_as_list()

    currency_crosses.drop(columns=["tag", "id"], inplace=True)
    currency_crosses = currency_crosses.where(pd.notnull(currency_crosses), None)

    if base is None and second is None:
        currency_crosses.reset_index(drop=True, inplace=True)
        return currency_crosses
    elif base is not None:
        base = unidecode(base.strip().upper())

        if base not in available_currencies:
            raise ValueError(
                "ERR#0053: the introduced currency " + base + " does not exists."
            )

        if second is not None:
            second = unidecode(second.strip().upper())

            if second not in available_currencies:
                raise ValueError(
                    "ERR#0053: the introduced currency " + second + " does not exists."
                )

            currency_crosses = currency_crosses[
                (currency_crosses["base"] == base)
                & (currency_crosses["second"] == second)
            ]

            currency_crosses.reset_index(drop=True, inplace=True)

            if len(currency_crosses) > 0:
                return currency_crosses
            else:
                raise ValueError(
                    "ERR#0054: the introduced currency cross "
                    + base
                    + "/"
                    + second
                    + " does not exists."
                )
        else:
            currency_crosses = currency_crosses[currency_crosses["base"] == base]
            currency_crosses.reset_index(drop=True, inplace=True)

            return currency_crosses
    elif second is not None:
        second = unidecode(second.strip().upper())

        if second not in available_currencies:
            raise ValueError(
                "ERR#0053: the introduced currency " + second + " does not exists."
            )

        currency_crosses = currency_crosses[currency_crosses["second"] == second]
        currency_crosses.reset_index(drop=True, inplace=True)

        return currency_crosses


def currency_crosses_as_list(base=None, second=None):
    """
    This function retrieves all the available currency crosses from Investing.com and returns them as a
    :obj:`dict`, which contains not just the currency crosses names, but all the fields contained on
    the currency_crosses file is columns is None, otherwise, just the specified column values will be returned. Note
    that the filtering params are both base and second, which mean the base and the second currency of the currency
    cross, for example, in the currency cross `EUR/USD` the base currency is EUR and the second currency is USD. These
    are optional parameters, so specifying one of them means that all the currency crosses where the introduced
    currency is either base or second will be returned; if both are specified, just the introduced currency cross will
    be returned if it exists. All the available currency crosses can be found at: https://www.investing.com/currencies/

    Args:
        base (:obj:`str`, optional):
            symbol of the base currency of the currency cross, this will return a :obj:`pandas.DataFrame` containing
            all the currency crosses where the base currency matches the introduced one.
        second (:obj:`str`):
            symbol of the second currency of the currency cross, this will return a :obj:`pandas.DataFrame` containing
            all the currency crosses where the second currency matches the introduced one.

    Returns:
        :obj:`list` - currency_crosses_list:
            The resulting :obj:`list` contains the retrieved data from the `currency_crosses.csv` file, which is
            a listing of the names of the currency crosses listed in Investing.com, which is the input for data
            retrieval functions as the name of the currency cross to retrieve data from needs to be specified.

            In case the listing was successfully retrieved, the :obj:`list` will look like::

                currency_crosses_list = [
                    'USD/BRLT', 'CAD/CHF', 'CHF/CAD', 'CAD/PLN', 'PLN/CAD', ...
                ]

    Raises:
        ValueError: raised if any of the introduced arguments is not valid or errored.
        FileNotFoundError: raised if `currency_crosses.csv` file was not found.
        IOError: raised if currency crosses retrieval failed, both for missing file or empty file.

    """

    if base is not None and not isinstance(base, str):
        raise ValueError("ERR#0049: specified base currency value is not valid.")

    if second is not None and not isinstance(second, str):
        raise ValueError("ERR#0051: specified second currency value is not valid.")

    resource_package = "investpy"
    resource_path = "/".join(("resources", "currency_crosses.csv"))
    if pkg_resources.resource_exists(resource_package, resource_path):
        currency_crosses = pd.read_csv(
            pkg_resources.resource_filename(resource_package, resource_path),
            keep_default_na=False,
        )
    else:
        raise FileNotFoundError("ERR#0060: currency_crosses file not found or errored.")

    if currency_crosses is None:
        raise IOError("ERR#0050: currency_crosses not found or unable to retrieve.")

    available_currencies = available_currencies_as_list()

    currency_crosses.drop(columns=["tag", "id"], inplace=True)
    currency_crosses = currency_crosses.where(pd.notnull(currency_crosses), None)

    if base is None and second is None:
        currency_crosses.reset_index(drop=True, inplace=True)
        return currency_crosses["name"].tolist()
    elif base is not None:
        base = unidecode(base.strip().upper())

        if base not in available_currencies:
            raise ValueError(
                "ERR#0053: the introduced currency " + base + " does not exists."
            )

        if second is not None:
            second = unidecode(second.strip().upper())

            if second not in available_currencies:
                raise ValueError(
                    "ERR#0053: the introduced currency " + second + " does not exists."
                )

            currency_crosses = currency_crosses[
                (currency_crosses["base"] == base)
                & (currency_crosses["second"] == second)
            ]

            currency_crosses.reset_index(drop=True, inplace=True)

            if len(currency_crosses) > 0:
                return currency_crosses["name"].tolist()
            else:
                raise ValueError(
                    "ERR#0054: the introduced currency cross "
                    + base
                    + "/"
                    + second
                    + " does not exists."
                )
        else:
            currency_crosses = currency_crosses[currency_crosses["base"] == base]
            currency_crosses.reset_index(drop=True, inplace=True)

            return currency_crosses["name"].tolist()
    elif second is not None:
        second = unidecode(second.strip().upper())

        if second not in available_currencies:
            raise ValueError(
                "ERR#0053: the introduced currency " + second + " does not exists."
            )

        currency_crosses = currency_crosses[currency_crosses["second"] == second]
        currency_crosses.reset_index(drop=True, inplace=True)

        return currency_crosses["name"].tolist()


def currency_crosses_as_dict(base=None, second=None, columns=None, as_json=False):
    """
    This function retrieves all the available currency crosses from Investing.com and returns them as a
    :obj:`dict`, which contains not just the currency crosses names, but all the fields contained on
    the currency_crosses file is columns is None, otherwise, just the specified column values will be returned. Note
    that the filtering params are both base and second, which mean the base and the second currency of the currency
    cross, for example, in the currency cross `EUR/USD` the base currency is EUR and the second currency is USD. These
    are optional parameters, so specifying one of them means that all the currency crosses where the introduced
    currency is either base or second will be returned; if both are specified, just the introduced currency cross will
    be returned if it exists. All the available currency crosses can be found at: https://www.investing.com/currencies/

    Args:
        base (:obj:`str`, optional):
            symbol of the base currency of the currency cross, this will return a :obj:`pandas.DataFrame` containing
            all the currency crosses where the base currency matches the introduced one.
        second (:obj:`str`):
            symbol of the second currency of the currency cross, this will return a :obj:`pandas.DataFrame` containing
            all the currency crosses where the second currency matches the introduced one.
        columns (:obj:`list`, optional):
            names of the columns of the currency crosses data to retrieve <name, full_name, base, base_name,
            second, second_name>
        as_json (:obj:`bool`, optional):
            value to determine the format of the output data which can either be a :obj:`dict` or a :obj:`json`.

    Returns:
        :obj:`dict` or :obj:`json` - currency_crosses_dict:
            The resulting :obj:`dict` contains the retrieved data if found, if not, the corresponding
            fields are filled with `None` values.

            In case the information was successfully retrieved, the :obj:`dict` will look like::

                currency_crosses_dict = {
                    'name': name,
                    'full_name': full_name,
                    'base': base,
                    'base_name': base_name,
                    'second': second,
                    'second_name': second_name
                }

    Raises:
        ValueError: raised if any of the introduced arguments is not valid or errored.
        FileNotFoundError: raised if `currency_crosses.csv` file was not found.
        IOError: raised if currency crosses retrieval failed, both for missing file or empty file.

    """

    if base is not None and not isinstance(base, str):
        raise ValueError("ERR#0049: specified base currency value is not valid.")

    if second is not None and not isinstance(second, str):
        raise ValueError("ERR#0051: specified second currency value is not valid.")

    if not isinstance(as_json, bool):
        raise ValueError(
            "ERR#0002: as_json argument can just be True or False, bool type."
        )

    resource_package = "investpy"
    resource_path = "/".join(("resources", "currency_crosses.csv"))
    if pkg_resources.resource_exists(resource_package, resource_path):
        currency_crosses = pd.read_csv(
            pkg_resources.resource_filename(resource_package, resource_path),
            keep_default_na=False,
        )
    else:
        raise FileNotFoundError("ERR#0060: currency_crosses file not found or errored.")

    if currency_crosses is None:
        raise IOError("ERR#0050: currency_crosses not found or unable to retrieve.")

    available_currencies = available_currencies_as_list()

    currency_crosses.drop(columns=["tag", "id"], inplace=True)
    currency_crosses = currency_crosses.where(pd.notnull(currency_crosses), None)

    if columns is None:
        columns = currency_crosses.columns.tolist()
    else:
        if not isinstance(columns, list):
            raise ValueError(
                "ERR#0020: specified columns argument is not a list, it can just be"
                " list type."
            )

    if not all(column in currency_crosses.columns.tolist() for column in columns):
        raise ValueError(
            "ERR#0021: specified columns does not exist, available columns are "
            "<name, full_name, base, base_name, second, second_name>"
        )

    if base is None and second is None:
        currency_crosses.reset_index(drop=True, inplace=True)

        if as_json:
            return json.dumps(currency_crosses[columns].to_dict(orient="records"))
        else:
            return currency_crosses[columns].to_dict(orient="records")
    elif base is not None:
        base = unidecode(base.strip().upper())

        if base not in available_currencies:
            raise ValueError(
                "ERR#0053: the introduced currency " + base + " does not exists."
            )

        if second is not None:
            second = unidecode(second.strip().upper())

            if second not in available_currencies:
                raise ValueError(
                    "ERR#0053: the introduced currency " + second + " does not exists."
                )

            currency_crosses = currency_crosses[
                (currency_crosses["base"] == base)
                & (currency_crosses["second"] == second)
            ]

            currency_crosses.reset_index(drop=True, inplace=True)

            if len(currency_crosses) > 0:
                if as_json:
                    return json.dumps(
                        currency_crosses[columns].to_dict(orient="records")
                    )
                else:
                    return currency_crosses[columns].to_dict(orient="records")
            else:
                raise ValueError(
                    "ERR#0054: the introduced currency cross "
                    + base
                    + "/"
                    + second
                    + " does not exists."
                )
        else:
            currency_crosses = currency_crosses[currency_crosses["base"] == base]
            currency_crosses.reset_index(drop=True, inplace=True)

            if as_json:
                return json.dumps(currency_crosses[columns].to_dict(orient="records"))
            else:
                return currency_crosses[columns].to_dict(orient="records")
    elif second is not None:
        second = unidecode(second.strip().upper())

        if second not in available_currencies:
            raise ValueError(
                "ERR#0053: the introduced currency " + second + " does not exists."
            )

        currency_crosses = currency_crosses[
            currency_crosses["second"] == unidecode(second.upper())
        ]
        currency_crosses.reset_index(drop=True, inplace=True)

        if as_json:
            return json.dumps(currency_crosses[columns].to_dict(orient="records"))
        else:
            return currency_crosses[columns].to_dict(orient="records")


def available_currencies_as_list():
    """
    This function retrieves a listing with all the available currencies with indexed currency crosses in order to
    get to know which are the available currencies. The currencies listed in this function, so on, can be used to
    search currency crosses and used the retrieved data to get historical data of those currency crosses, so to
    determine which is the value of one base currency in the second currency.

    Returns:
        :obj:`list` - available_currencies:
            The resulting :obj:`list` contains all the available currencies with currency crosses being either the base
            or the second value of the cross, as listed in Investing.com.

    """

    return list(cst.CURRENCIES.keys())



# Copyright 2018-2021 Alvaro Bartolome, alvarobartt @ GitHub
# See LICENSE for details.

import json
from datetime import date, datetime, timedelta
from random import randint

import pandas as pd
import pytz
import requests
from lxml.html import fromstring

from .constant import FUNDS_INTERVAL_FILTERS, INTERVAL_FILTERS, OUTDATED2UPDATED
from .data import Data
from .extra import random_user_agent


class SearchObj(object):
    """Class which contains each search result when searching data in Investing.com.

    This class contains the search results of the Investing.com search made with the function
    call `investpy.search_quotes(text, products, countries, n_results)` which returns a :obj:`list`
    of instances of this class with the formatted retrieved information. Additionally, data can
    either be retrieved or not including both recent and historical data, which will be included
    in the `SearchObj.data` attribute when calling either `SearchObj.retrieve_recent_data()` or
    `SearchObj.retrieve_historical_data(from_date, to_date)`, respectively.

    Attributes:
        id_ (:obj:`int`): ID value used by Investing.com to retrieve data.
        name (:obj:`str`): name of the retrieved financial product.
        symbol (:obj:`str`): symbol of the retrieved financial product.
        tag (:obj:`str`): tag (which is the Investing.com URL) of the retrieved financial product.
        country (:obj:`str`): name of the country from where the retrieved financial product is.
        pair_type (:obj:`str`): type of retrieved financial product (stocks, funds, etfs, etc.).
        exchange (:obj:`str`): name of the stock exchange of the retrieved financial product.

    Extra Attributes:
        data (:obj:`pandas.DataFrame`):
            recent or historical data to retrieve from the current financial product, generated
            after calling either self.retrieve_recent_data or self.retrieve_historical_data().
        info (:obj:`dict`):
            contains the information of the current financial product, generated after calling the
            self.retrieve_information() function.

    """

    def __init__(self, id_, name, symbol, tag, country, pair_type, exchange):
        self.id_ = id_
        self.name = name
        self.symbol = symbol
        self.country = country
        self.tag = tag
        self.pair_type = pair_type
        self.exchange = exchange

    def __str__(self):
        return json.dumps(self.__dict__)

    def __eq__(self, other):
        return self.id_ == other.id_

    def __hash__(self):
        return self.id_

    def retrieve_recent_data(self):
        """Class method used to retrieve the recent data from the class instance of any financial product.

        This method retrieves the recent data from Investing.com of the financial product of the current class
        instance, so it fills the `SearchObj.data` attribute with the retrieved :obj:`pandas.DataFrame`. This method
        uses the previously filled data from the `investpy.search_quotes(text, products, countries, n_results)` function
        search results to build the request that it is going to be sent to Investing.com so to retrieve and parse the data.

        Returns:
            :obj:`pandas.DataFrame` - data:
                This method retrieves the recent data from the current class instance of a financial product
                from Investing.com. This method both stores retrieved data in self.data attribute of the class
                instance and it also returns it as a normal function will do.

        """

        if self.pair_type in ["stocks", "funds", "etfs", "currencies", "certificates"]:
            header = f"{self.symbol} Historical Data"
            headers, params = self._prepare_request(header)
        elif self.pair_type in ["bonds"]:
            header = f"{self.name} Bond Yield Historical Data"
            headers, params = self._prepare_request(header)
        elif self.pair_type in ["indices", "commodities", "cryptos", "fxfutures"]:
            header = f"{self.name} Historical Data"
            headers, params = self._prepare_request(header)

        self.data = self._data_retrieval(
            product=self.pair_type, headers=headers, params=params
        )
        self._convert2df()
        return self.data

    def retrieve_historical_data(self, from_date, to_date):
        """Class method used to retrieve the historical data from the class instance of any financial product.

        This method retrieves the historical data from Investing.com of the financial product of the current class
        instance on the specified date range, so it fills the `SearchObj.data` attribute with the retrieved
        :obj:`pandas.DataFrame`. This method uses the previously filled data from the
        `investpy.search_quotes(text, products, countries, n_results)` function search results to build the request
        that it is going to be sent to Investing.com so to retrieve and parse the data.

        Args:
            from_date (:obj:`str`): date from which data will be retrieved, specified in dd/mm/yyyy format.
            to_date (:obj:`str`): date until data will be retrieved, specified in dd/mm/yyyy format.

        Returns:
            :obj:`pandas.DataFrame` - data:
                This method retrieves the historical data from the current class instance of a financial product
                from Investing.com. This method both stores retrieved data in self.data attribute of the class
                instance and it also returns it as a normal function will do.

        Raises:
            ValueError: raised if any of the introduced parameters was not valid or errored.
            RuntimeError: raised if there was any error while retrieving the data from Investing.com.

        """

        try:
            datetime.strptime(from_date, "%d/%m/%Y")
        except ValueError:
            raise ValueError(
                "ERR#0011: incorrect from_date date format, it should be 'dd/mm/yyyy'."
            )

        try:
            datetime.strptime(to_date, "%d/%m/%Y")
        except ValueError:
            raise ValueError(
                "ERR#0012: incorrect to_date format, it should be 'dd/mm/yyyy'."
            )

        from_date = datetime.strptime(from_date, "%d/%m/%Y")
        to_date = datetime.strptime(to_date, "%d/%m/%Y")

        if from_date >= to_date:
            raise ValueError(
                "ERR#0032: to_date should be greater than from_date, both formatted as"
                " 'dd/mm/yyyy'."
            )

        if self.pair_type in ["stocks", "funds", "etfs", "currencies", "certificates"]:
            header = f"{self.symbol} Historical Data"
        elif self.pair_type in ["bonds"]:
            header = f"{self.name} Bond Yield Historical Data"
        elif self.pair_type in ["indices", "commodities", "cryptos", "fxfutures"]:
            header = f"{self.name} Historical Data"

        if to_date.year - from_date.year > 19:
            intervals = self._calculate_intervals(from_date, to_date)

            self.data = list()

            for interval in intervals:
                headers, params = self._prepare_historical_request(
                    header=header, from_date=interval["from"], to_date=interval["to"]
                )
                try:
                    self.data += self._data_retrieval(
                        product=self.pair_type, headers=headers, params=params
                    )
                except:
                    continue

            if len(self.data) < 1:
                raise IndexError("ERR#0033: information unavailable or not found.")
        else:
            headers, params = self._prepare_historical_request(
                header=header,
                from_date=from_date.strftime("%m/%d/%Y"),
                to_date=to_date.strftime("%m/%d/%Y"),
            )
            self.data = self._data_retrieval(
                product=self.pair_type, headers=headers, params=params
            )

        self._convert2df()
        return self.data

    def retrieve_information(self):
        """Class method used to retrieve the information from the class instance of any financial product.

        This method retrieves the information from Investing.com of the financial product of the current class
        instance, so it fills the `SearchObj.info` attribute with the retrieved :obj:`dict`. This method uses the
        previously retrieved data from the `investpy.search_quotes(text, products, countries, n_results)`
        function search results to build the request that it is going to be sent to Investing.com so to retrieve and
        parse the information, since the product tag is required.

        Returns:
            :obj:`dict` - info:
                This method retrieves the information from the current class instance of a financial product
                from Investing.com. This method both stores retrieved information in self.information attribute of the class
                instance and it also returns it as a normal function will do.

        Raises:
            ConnectionError: raised if connection to Investing.com could not be established.
            RuntimeError: raised if there was any problem while retrieving the data from Investing.com.

        """

        url = f"https://www.investing.com{self.tag}"

        headers = {
            "User-Agent": random_user_agent(),
            "X-Requested-With": "XMLHttpRequest",
            "Accept": "text/html",
            "Accept-Encoding": "gzip, deflate",
            "Connection": "keep-alive",
        }

        req = requests.get(url, headers=headers)

        if req.status_code != 200:
            raise ConnectionError(
                f"ERR#0015: error {req.status_code}, try again later."
            )

        # Just change this list once the update is included for all the other products
        updated_for = ["stocks"]
        outdated_for = [
            "etfs",
            "commodities",
            "currencies",
            "funds",
            "bonds",
            "cryptos",
            "certificates",
            "indices",
            "fxfutures",
        ]

        root_ = fromstring(req.text)
        updated_path = root_.xpath("//dl[@data-test='key-info']/div")
        outdated_path = root_.xpath("//div[contains(@class, 'overviewDataTable')]/div")

        if not updated_path and not outdated_path:
            raise RuntimeError("ERR#0004: data retrieval error while scraping.")

        path_, investing_updated = (
            (updated_path, True) if updated_path else (outdated_path, False)
        )

        self.information = dict()

        for elements_ in path_:
            if investing_updated:
                element = elements_.xpath(".//dd")[0]
                title = element.get("data-test")
            else:
                element = elements_.xpath(".//span[@class='float_lang_base_1']")[0]
                title = element.text_content().strip()
                title = OUTDATED2UPDATED[title]
                element = element.getnext()
            try:
                value = float(element.text_content().replace(",", ""))
                if isinstance(value, float):
                    if value.is_integer() is True:
                        value = int(value)
                self.information[title] = value if value != "N/A" else None
                continue
            except:
                pass
            try:
                text = element.text_content().strip()
                in_format = "%b %d, %Y" if investing_updated else "%m/%d/%Y"
                text = datetime.strptime(text, in_format).strftime("%d/%m/%Y")
                self.information[title] = text if text != "N/A" else None
                continue
            except:
                pass
            try:
                text = element.text_content().strip()
                if text.__contains__("1 = "):
                    text = text.replace("1 = ", "")
                    self.information[title] = text if text != "N/A" else None
                    continue
            except:
                pass
            try:
                value = element.text_content().strip()
                if value.__contains__("K"):
                    value = float(value.replace("K", "").replace(",", "")) * 1e3
                elif value.__contains__("M"):
                    value = float(value.replace("M", "").replace(",", "")) * 1e6
                elif value.__contains__("B"):
                    value = float(value.replace("B", "").replace(",", "")) * 1e9
                elif value.__contains__("T"):
                    value = float(value.replace("T", "").replace(",", "")) * 1e12
                if isinstance(value, float):
                    if value.is_integer() is True:
                        value = int(value)
                self.information[title] = value if value != "N/A" else None
                continue
            except:
                pass

        return self.information

    def retrieve_technical_indicators(self, interval="daily"):
        """Class method used to retrieve the technical indicators from the class instance of any financial product.

        This method retrieves the technical indicators from Investing.com for the financial product of the current
        class instance, to later put in into the `SearchObj.technical_indicators` attribute. This method uses the
        previously retrieved data from the `investpy.search_quotes(text, products, countries, n_results)`
        function search results to build the request that it is going to be sent to Investing.com so to retrieve and
        parse the technical indicators, since the product id is required.

        Args:
            interval (:obj:`str`, optional):
                time interval of the technical indicators' calculations, available values are: `5mins`, `15mins`,
                `30mins`, `1hour`, `5hours`, `daily`, `weekly` and `monthly`. Note that for funds just the intervals:
                `daily`, `weekly` and `monthly` are available.

        Returns:
            :obj:`pd.DataFrame` - technical_indicators:
                This method retrieves the technical indicators from the current class instance of a financial product
                from Investing.com. This method not just stores retrieved technical indicators table into self.technical_indicators
                but it also returns it as a normal function will do.

        Raises:
            ValueError: raised if any of the input parameters is not valid.
            ConnectionError: raised if connection to Investing.com could not be established.
            RuntimeError: raised if there was any problem while retrieving the data from Investing.com.

        """

        if self.pair_type in []:
            raise ValueError(
                "Investing.com does not provide technical indicators for"
                f" {self.pair_type}."
            )

        if self.pair_type != "funds" and interval not in INTERVAL_FILTERS:
            raise ValueError(
                "Investing.com just provides the following intervals for"
                f" {self.pair_type}' technical indicators:"
                f" {', '.join(list(INTERVAL_FILTERS.keys()))}"
            )

        if self.pair_type == "funds" and interval not in FUNDS_INTERVAL_FILTERS:
            raise ValueError(
                "Investing.com just provides the following intervals for funds'"
                " technical indicators:"
                f" {', '.join(list(FUNDS_INTERVAL_FILTERS.keys()))}"
            )

        params = {
            "pairID": self.id_,
            "period": INTERVAL_FILTERS[interval],
            "viewType": "normal",
        }

        headers = {
            "User-Agent": random_user_agent(),
            "X-Requested-With": "XMLHttpRequest",
            "Accept": "text/html",
            "Accept-Encoding": "gzip, deflate",
            "Connection": "keep-alive",
        }

        url = "https://www.investing.com/instruments/Service/GetTechincalData"

        req = requests.post(url, headers=headers, data=params)

        if req.status_code != 200:
            raise ConnectionError(
                f"ERR#0015: error {req.status_code}, try again later."
            )

        root_ = fromstring(req.text)
        table_ = root_.xpath(
            ".//table[contains(@class, 'technicalIndicatorsTbl')]/tbody/tr"
        )

        if not table_:
            raise RuntimeError("ERR#0004: data retrieval error while scraping.")

        self.technical_indicators = pd.DataFrame()

        for row in table_:
            for value in row.xpath("td"):
                if value.get("class").__contains__("symbol"):
                    self.technical_indicators = self.technical_indicators.append(
                        {
                            "indicator": value.text_content().strip(),
                            "value": float(value.getnext().text_content().strip()),
                            "signal": (
                                value.getnext().getnext().text_content().strip().lower()
                            ).replace(" ", "_"),
                        },
                        ignore_index=True,
                    )

        return self.technical_indicators

    def retrieve_currency(self):
        """Class method used to retrieve the default currency from the class instance of any financial product.

        This method retrieves the default currency from Investing.com of the financial product of the current class
        instance. This method uses the data previously retrieved from the `investpy.search_quotes(text, products, countries, n_results)`
        function search results to build the request that it is going to be sent to Investing.com so to retrieve and
        parse the information, since the product tag is required.

        Returns:
            :obj:`str` - default_currency:
                This method retrieves the default currency from the current class instance of a financial product
                from Investing.com.

        Raises:
            ConnectionError: raised if connection to Investing.com could not be established.
            RuntimeError: raised if there was any problem while retrieving the data from Investing.com.

        """

        url = f"https://www.investing.com{self.tag}"

        headers = {
            "User-Agent": random_user_agent(),
            "X-Requested-With": "XMLHttpRequest",
            "Accept": "text/html",
            "Accept-Encoding": "gzip, deflate",
            "Connection": "keep-alive",
        }

        req = requests.get(url, headers=headers)

        if req.status_code != 200:
            raise ConnectionError(
                f"ERR#0015: error {req.status_code}, try again later."
            )

        # Just change this list once the update is included for all the other products
        updated_for = ["stocks"]
        outdated_for = [
            "etfs",
            "commodities",
            "currencies",
            "funds",
            "bonds",
            "cryptos",
            "certificates",
            "indices",
            "fxfutures",
        ]

        root_ = fromstring(req.text)
        updated_path = root_.xpath(
            "//div[contains(@class, 'instrument-metadata_currency')]/span"
        )
        outdated_path = root_.xpath(
            "//div[@id='quotes_summary_current_data']/div/div/div[contains(@class,"
            " 'bottom')]/span[@class='bold']"
        )

        if not updated_path and not outdated_path:
            raise RuntimeError("ERR#0004: data retrieval error while scraping.")

        path_, investing_updated = (
            (updated_path, True) if updated_path else (outdated_path, False)
        )

        self.default_currency = path_[-1].text_content().strip()
        return self.default_currency

    def _prepare_request(self, header):
        headers = {
            "User-Agent": random_user_agent(),
            "X-Requested-With": "XMLHttpRequest",
            "Accept": "text/html",
            "Accept-Encoding": "gzip, deflate",
            "Connection": "keep-alive",
        }

        params = {
            "curr_id": self.id_,
            "smlID": str(randint(1000000, 99999999)),
            "header": header,
            "interval_sec": "Daily",
            "sort_col": "date",
            "sort_ord": "DESC",
            "action": "historical_data",
        }

        return headers, params

    def _prepare_historical_request(self, header, from_date, to_date):
        headers = {
            "User-Agent": random_user_agent(),
            "X-Requested-With": "XMLHttpRequest",
            "Accept": "text/html",
            "Accept-Encoding": "gzip, deflate",
            "Connection": "keep-alive",
        }

        params = {
            "curr_id": self.id_,
            "smlID": str(randint(1000000, 99999999)),
            "header": header,
            "st_date": from_date,
            "end_date": to_date,
            "interval_sec": "Daily",
            "sort_col": "date",
            "sort_ord": "DESC",
            "action": "historical_data",
        }

        return headers, params

    def _calculate_intervals(self, from_date, to_date):
        intervals = list()

        while True:
            diff = to_date.year - from_date.year

            if diff > 19:
                intervals.append(
                    {
                        "from": from_date.strftime("%m/%d/%Y"),
                        "to": from_date.replace(year=from_date.year + 19).strftime(
                            "%m/%d/%Y"
                        ),
                    }
                )

                from_date = from_date.replace(year=from_date.year + 19) + timedelta(
                    days=1
                )
            else:
                intervals.append(
                    {
                        "from": from_date.strftime("%m/%d/%Y"),
                        "to": to_date.strftime("%m/%d/%Y"),
                    }
                )

                break

        return intervals

    def _data_retrieval(self, product, headers, params):
        has_volume = (
            True
            if product
            in ["stocks", "etfs", "indices", "cryptos", "commodities", "fxfutures"]
            else False
        )
        has_change_pct = True  # Every financial product has it

        url = "https://www.investing.com/instruments/HistoricalDataAjax"

        req = requests.post(url, headers=headers, data=params)

        if req.status_code != 200:
            raise ConnectionError(
                f"ERR#0015: error {req.status_code}, try again later."
            )

        root_ = fromstring(req.text)
        path_ = root_.xpath(".//table[@id='curr_table']/tbody/tr")

        if not path_:
            raise RuntimeError("ERR#0004: data retrieval error while scraping.")

        results = list()

        for elements_ in path_:
            info = []

            for nested_ in elements_.xpath(".//td"):
                val = (
                    nested_.get("data-real-value")
                    if nested_.get("data-real-value") is not None
                    else nested_.text_content()
                )
                if val == "No results found":
                    raise IndexError("ERR#0033: information unavailable or not found.")
                info.append(val)

            result = {
                "Date": datetime.strptime(
                    str(
                        datetime.fromtimestamp(
                            int(info[0]), tz=pytz.timezone("GMT")
                        ).date()
                    ),
                    "%Y-%m-%d",
                ),
                "Open": float(info[2].replace(",", "")),
                "High": float(info[3].replace(",", "")),
                "Low": float(info[4].replace(",", "")),
                "Close": float(info[1].replace(",", "")),
            }

            if has_volume and has_change_pct:
                result["Volume"] = int(info[5])
                result["Change Pct"] = float(info[6].replace(",", "").replace("%", ""))

            if not has_volume and has_change_pct:
                result["Change Pct"] = float(info[5].replace(",", "").replace("%", ""))

            results.append(result)

        return results[::-1]

    def _convert2df(self):
        self.data = pd.DataFrame(self.data)
        self.data.set_index("Date", inplace=True)



# Copyright 2018-2021 Alvaro Bartolome, alvarobartt @ GitHub
# See LICENSE for details.

import json

import pandas as pd
import pkg_resources
from unidecode import unidecode

from ..utils import constant as cst


def funds_as_df(country=None):
    """
    This function retrieves all the available `funds` from Investing.com and returns them as a :obj:`pandas.DataFrame`,
    which contains not just the fund names, but all the fields contained on the `funds.csv` file.
    All the available funds can be found at: https://www.investing.com/funds/

    Args:
        country (:obj:`str`, optional): name of the country to retrieve all its available funds from.

    Returns:
        :obj:`pandas.DataFrame` - funds_df:
            The resulting :obj:`pandas.DataFrame` contains all the funds basic information retrieved from Investing.com,
            some of which is not useful for the user, but for the inner package functions, such as the `id` field,
            for example.

            In case the information was successfully retrieved, the :obj:`pandas.DataFrame` will look like::

                country | name | symbol | issuer | isin | asset_class | currency | underlying
                --------|------|--------|--------|------|-------------|----------|------------
                xxxxxxx | xxxx | xxxxxx | xxxxxx | xxxx | xxxxxxxxxxx | xxxxxxxx | xxxxxxxxxx

    Raises:
        ValueError: raised whenever any of the introduced arguments is not valid or errored.
        FileNotFoundError: raised when the `funds.csv` file was not found.
        IOError: raised if the `funds.csv` file is missing or errored.

    """

    if country is not None and not isinstance(country, str):
        raise ValueError("ERR#0025: specified country value not valid.")

    resource_package = "investpy"
    resource_path = "/".join(("resources", "funds.csv"))
    if pkg_resources.resource_exists(resource_package, resource_path):
        funds = pd.read_csv(
            pkg_resources.resource_filename(resource_package, resource_path),
            keep_default_na=False,
        )
    else:
        raise FileNotFoundError("ERR#0057: funds file not found or errored.")

    if funds is None:
        raise IOError("ERR#0005: funds not found or unable to retrieve.")

    funds.drop(columns=["tag", "id"], inplace=True)
    funds = funds.where(pd.notnull(funds), None)

    if country is None:
        funds.reset_index(drop=True, inplace=True)
        return funds
    else:
        country = unidecode(country.strip().lower())

        if country not in fund_countries_as_list():
            raise ValueError(
                "ERR#0034: country " + country + " not found, check if it is correct."
            )

        funds = funds[funds["country"] == unidecode(country.lower())]
        funds.reset_index(drop=True, inplace=True)

        return funds


def funds_as_list(country=None):
    """
    This function retrieves all the available funds and returns a list of each one of them.
    All the available funds can be found at: https://www.investing.com/funds/

    Args:
        country (:obj:`str`, optional): name of the country to retrieve all its available funds from.

    Returns:
        :obj:`list` - funds_list:
            The resulting list contains the retrieved data, which corresponds to the fund names of
            every fund listed on Investing.com.

            In case the information was successfully retrieved from the CSV file, the :obj:`list` will look like::

                funds = [
                    'Blackrock Global Funds - Global Allocation Fund E2',
                    'Quality Inversión Conservadora Fi',
                    'Nordea 1 - Stable Return Fund E Eur',
                    ...
                ]

    Raises:
        ValueError: raised whenever any of the introduced arguments is not valid or errored.
        FileNotFoundError: raised when the `funds.csv` file was not found.
        IOError: raised if the `funds.csv` file is missing or errored.

    """

    if country is not None and not isinstance(country, str):
        raise ValueError("ERR#0025: specified country value not valid.")

    resource_package = "investpy"
    resource_path = "/".join(("resources", "funds.csv"))
    if pkg_resources.resource_exists(resource_package, resource_path):
        funds = pd.read_csv(
            pkg_resources.resource_filename(resource_package, resource_path),
            keep_default_na=False,
        )
    else:
        raise FileNotFoundError("ERR#0057: funds file not found or errored.")

    if funds is None:
        raise IOError("ERR#0005: funds not found or unable to retrieve.")

    funds.drop(columns=["tag", "id"], inplace=True)
    funds = funds.where(pd.notnull(funds), None)

    if country is None:
        return funds["name"].tolist()
    else:
        country = unidecode(country.strip().lower())

        if country not in fund_countries_as_list():
            raise ValueError(
                "ERR#0034: country " + country + " not found, check if it is correct."
            )

        return funds[funds["country"] == country]["name"].tolist()


def funds_as_dict(country=None, columns=None, as_json=False):
    """
    This function retrieves all the available funds on Investing.com and returns them as a :obj:`dict` containing
    the country, name, symbol, tag, id, issuer, isin, asset_class, currency and underlying data. All the available
    funds can be found at: https://www.investing.com/funds/

    Args:
        country (:obj:`str`, optional): name of the country to retrieve all its available funds from.
        columns (:obj:`list` of :obj:`str`, optional): description
            a :obj:`list` containing the column names from which the data is going to be retrieved.
        as_json (:obj:`bool`, optional): description
            value to determine the format of the output data (:obj:`dict` or :obj:`json`).

    Returns:
        :obj:`dict` or :obj:`json` - funds_dict:
            The resulting :obj:`dict` contains the retrieved data if found, if not, the corresponding
            fields are filled with `None` values.

            In case the information was successfully retrieved, the :obj:`dict` will look like::

                {
                    'country': country,
                    'name': name,
                    'symbol': symbol,
                    'issuer': issuer,
                    'isin': isin,
                    'asset_class': asset_class,
                    'currency': currency,
                    'underlying': underlying
                }

    Raises:
        ValueError: raised whenever any of the introduced arguments is not valid or errored.
        FileNotFoundError: raised when the `funds.csv` file was not found.
        IOError: raised if the `funds.csv` file is missing or errored.

    """

    if country is not None and not isinstance(country, str):
        raise ValueError("ERR#0025: specified country value not valid.")

    if not isinstance(as_json, bool):
        raise ValueError(
            "ERR#0002: as_json argument can just be True or False, bool type."
        )

    resource_package = "investpy"
    resource_path = "/".join(("resources", "funds.csv"))
    if pkg_resources.resource_exists(resource_package, resource_path):
        funds = pd.read_csv(
            pkg_resources.resource_filename(resource_package, resource_path),
            keep_default_na=False,
        )
    else:
        raise FileNotFoundError("ERR#0057: funds file not found or errored.")

    if funds is None:
        raise IOError("ERR#0005: funds not found or unable to retrieve.")

    funds.drop(columns=["tag", "id"], inplace=True)
    funds = funds.where(pd.notnull(funds), None)

    if columns is None:
        columns = funds.columns.tolist()
    else:
        if not isinstance(columns, list):
            raise ValueError(
                "ERR#0020: specified columns argument is not a list, it can just be"
                " list type."
            )

    if not all(column in funds.columns.tolist() for column in columns):
        raise ValueError(
            "ERR#0023: specified columns does not exist, available columns are "
            "<country, name, symbol, issuer, isin, asset_class, currency, underlying>"
        )

    if country is None:
        if as_json:
            return json.dumps(funds[columns].to_dict(orient="records"))
        else:
            return funds[columns].to_dict(orient="records")
    else:
        country = unidecode(country.strip().lower())

        if country not in fund_countries_as_list():
            raise ValueError(
                "ERR#0034: country " + country + " not found, check if it is correct."
            )

        if as_json:
            return json.dumps(
                funds[funds["country"] == country][columns].to_dict(orient="records")
            )
        else:
            return funds[funds["country"] == country][columns].to_dict(orient="records")


def fund_countries_as_list():
    """
    This function returns a listing with all the available countries from where funds can be retrieved, so to
    let the user know which of them are available, since the parameter country is mandatory in every fund retrieval
    function.

    Returns:
        :obj:`list` - countries:
            The resulting :obj:`list` contains all the available countries with funds as indexed in Investing.com

    """

    return [value["country"] for value in cst.FUND_COUNTRIES]
